from functools import wraps
from typing import TYPE_CHECKING, Any, Callable, Optional

import weave
from weave.trace.autopatch import OpSettings
from weave.trace.op import _add_accumulator
from weave.trace.serialization.serialize import dictify
from weave.trace.weave_client import Call

if TYPE_CHECKING:
    from google.genai.types import GenerateContentResponse

SKIP_TRACING_FUNCTIONS = [
    "google.genai.models.Models.count_tokens",
    "google.genai.models.AsyncModels.count_tokens",
]


def google_genai_gemini_postprocess_inputs(inputs: dict[str, Any]) -> dict[str, Any]:
    """
    Postprocess inputs of the trace for the Google GenAI Gemini API to be used in
    the trace visualization in the Weave UI. If the parameter `self` is present
    (i.e, if the function being traced is a stateful method), it is converted to a
    dictionary of attributes that can be displayed in the Weave UI.
    """
    # Extract the model name from the inputs and ensure it is present in the inputs
    model_name = getattr(inputs["self"], "_model", None)
    if model_name is not None:
        inputs["model"] = model_name

    # Convert the `self` parameter which is actually the state of the
    # `google.genai.models.Models` object to a dictionary of attributes that can
    # be displayed in the Weave UI
    if "self" in inputs:
        inputs["self"] = dictify(inputs["self"])
    return inputs


def google_genai_gemini_on_finish(
    call: Call, output: Any, exception: Optional[BaseException] = None
) -> None:
    """
    On finish handler for the Google GenAI Gemini API integration that ensures the usage
    metadata is added to the summary of the trace.
    """
    if not (model_name := call.inputs.get("model")):
        raise ValueError("Unknown model type")
    usage = {model_name: {"requests": 1}}
    summary_update = {"usage": usage}
    if output:
        call.output = dictify(output)
        if hasattr(output, "usage_metadata"):
            usage[model_name].update(
                {
                    "prompt_tokens": output.usage_metadata.prompt_token_count,
                    "completion_tokens": output.usage_metadata.candidates_token_count,
                    "total_tokens": output.usage_metadata.total_token_count,
                }
            )
    if call.summary is not None:
        call.summary.update(summary_update)


def google_genai_gemini_accumulator(
    acc: Optional["GenerateContentResponse"], value: "GenerateContentResponse"
) -> "GenerateContentResponse":
    print("google_genai_gemini_accumulator")
    if acc is None:
        return value

    for i, value_candidate in enumerate(value.candidates):
        if i >= len(acc.candidates):
            break
        for j, value_part in enumerate(value_candidate.content.parts):
            if j >= len(acc.candidates[i].content.parts):
                break
            if value_part.text is not None:
                acc.candidates[i].content.parts[j].text += value_part.text

    if acc.usage_metadata.prompt_token_count is None:
        acc.usage_metadata.prompt_token_count = 0
    elif value.usage_metadata.prompt_token_count is not None:
        acc.usage_metadata.prompt_token_count += value.usage_metadata.prompt_token_count

    if acc.usage_metadata.candidates_token_count is None:
        acc.usage_metadata.candidates_token_count = 0
    elif value.usage_metadata.candidates_token_count is not None:
        acc.usage_metadata.candidates_token_count += (
            value.usage_metadata.candidates_token_count
        )

    if acc.usage_metadata.total_token_count is None:
        acc.usage_metadata.total_token_count = 0
    elif value.usage_metadata.total_token_count is not None:
        acc.usage_metadata.total_token_count += value.usage_metadata.total_token_count

    if acc.usage_metadata.cached_content_token_count is None:
        acc.usage_metadata.cached_content_token_count = 0
    elif value.usage_metadata.cached_content_token_count is not None:
        acc.usage_metadata.cached_content_token_count += (
            value.usage_metadata.cached_content_token_count
        )

    return acc


def google_genai_gemini_wrapper_sync(
    settings: OpSettings,
) -> Callable[[Callable], Callable]:
    def wrapper(fn: Callable) -> Callable:
        op_kwargs = settings.model_dump()
        if not op_kwargs.get("postprocess_inputs"):
            op_kwargs["postprocess_inputs"] = google_genai_gemini_postprocess_inputs

        op = weave.op(fn, **op_kwargs)
        if op.name not in SKIP_TRACING_FUNCTIONS:
            op._set_on_finish_handler(google_genai_gemini_on_finish)
        return _add_accumulator(
            op,
            make_accumulator=lambda inputs: google_genai_gemini_accumulator,
            should_accumulate=lambda inputs: op.name.endswith("stream"),
        )

    return wrapper


def google_genai_gemini_wrapper_async(
    settings: OpSettings,
) -> Callable[[Callable], Callable]:
    def wrapper(fn: Callable) -> Callable:
        def _fn_wrapper(fn: Callable) -> Callable:
            @wraps(fn)
            async def _async_wrapper(*args: Any, **kwargs: Any) -> Any:
                return await fn(*args, **kwargs)

            return _async_wrapper

        op_kwargs = settings.model_dump()
        if not op_kwargs.get("postprocess_inputs"):
            op_kwargs["postprocess_inputs"] = google_genai_gemini_postprocess_inputs

        op = weave.op(_fn_wrapper(fn), **op_kwargs)
        if op.name not in SKIP_TRACING_FUNCTIONS:
            op._set_on_finish_handler(google_genai_gemini_on_finish)
        return _add_accumulator(
            op,
            make_accumulator=lambda inputs: google_genai_gemini_accumulator,
            should_accumulate=lambda inputs: op.name.endswith("stream"),
        )

    return wrapper



from weave.integrations.mcp.mcp_client import (
    get_mcp_client_patcher as get_mcp_client_patcher,
)
from weave.integrations.mcp.mcp_server import (
    get_mcp_server_patcher as get_mcp_server_patcher,
)



"""Progress bar utilities for WeaveClient.

This module provides functionality for displaying progress bars when flushing
tasks in the WeaveClient.
"""

from typing import Callable

from rich.console import Console
from rich.progress import (
    BarColumn,
    Progress,
    SpinnerColumn,
    TaskProgressColumn,
    TextColumn,
    TimeElapsedColumn,
)

from weave.trace.weave_client import FlushStatus


def create_progress_bar_callback() -> Callable[[FlushStatus], None]:
    """Create a callback function that displays a progress bar for flush status.

    Returns:
        A callback function that can be passed to WeaveClient._flush.
    """
    console = Console()

    # Create a progress bar instance
    progress = Progress(
        SpinnerColumn(),
        TextColumn("[bold blue]{task.description}"),
        BarColumn(bar_width=None, complete_style="magenta"),
        TaskProgressColumn(),
        TimeElapsedColumn(),
        console=console,
        refresh_per_second=10,
        expand=True,
        transient=False,
    )

    # Start the progress display
    progress.start()

    # Create a task for tracking progress
    task_id = None

    def progress_callback(status: FlushStatus) -> None:
        """Update the progress bar based on the flush status.

        Args:
            status: The current flush status.
        """
        nonlocal task_id

        counts = status["job_counts"]

        # If this is the first update, create the task
        if task_id is None:
            if counts["total_jobs"] == 0:
                # No jobs to track, just return
                progress.stop()
                return

            # Print initial message
            if not progress.live.is_started:
                print(f"Flushing {counts['total_jobs']} pending tasks...")

            # Create the task
            task_id = progress.add_task("Flushing tasks", total=counts["total_jobs"])

        # If there are no more pending jobs, complete the progress bar
        if not status["has_pending_jobs"]:
            progress.update(
                task_id,
                completed=status["max_total_jobs"],
                total=status["max_total_jobs"],
            )
            progress.stop()
            return

        # If new jobs were added, update the total
        if status["max_total_jobs"] > progress.tasks[task_id].total:
            progress.update(task_id, total=status["max_total_jobs"])

        # Update progress bar with completed jobs
        if status["completed_since_last_update"] > 0:
            progress.update(task_id, advance=status["completed_since_last_update"])

        # Format job details for description
        job_details = []
        if counts["main_jobs"] > 0:
            job_details.append(f"{counts['main_jobs']} main")
        if counts["fastlane_jobs"] > 0:
            job_details.append(f"{counts['fastlane_jobs']} file-upload")
        if counts["call_processor_jobs"] > 0:
            job_details.append(f"{counts['call_processor_jobs']} call-batch")

        job_details_str = ", ".join(job_details) if job_details else "none"

        # Update progress bar description
        progress.update(
            task_id,
            description=f"Flushing tasks ({counts['total_jobs']} remaining: {job_details_str})",
        )

    return progress_callback



from typing import Any

from pydantic import BaseModel, Field


class FeedbackPayloadReactionReq(BaseModel):
    emoji: str


class FeedbackPayloadNoteReq(BaseModel):
    note: str = Field(min_length=1, max_length=1024)


FEEDBACK_PAYLOAD_SCHEMAS: dict[str, type[BaseModel]] = {
    "wandb.reaction.1": FeedbackPayloadReactionReq,
    "wandb.note.1": FeedbackPayloadNoteReq,
}

ANNOTATION_FEEDBACK_TYPE_PREFIX = "wandb.annotation"
RUNNABLE_FEEDBACK_TYPE_PREFIX = "wandb.runnable"


# Making the decision to use `value` & `payload` as nested keys so that
# we can:
# 1. Add more fields in the future without breaking changes
# 2. Support primitive values for annotation feedback that still schema
class AnnotationPayloadSchema(BaseModel):
    value: Any


class RunnablePayloadSchema(BaseModel):
    output: Any


def feedback_type_is_annotation(feedback_type: str) -> bool:
    return feedback_type.startswith(ANNOTATION_FEEDBACK_TYPE_PREFIX)


def feedback_type_is_runnable(feedback_type: str) -> bool:
    return feedback_type.startswith(RUNNABLE_FEEDBACK_TYPE_PREFIX)


def runnable_feedback_selector(name: str) -> str:
    return f"feedback.[{RUNNABLE_FEEDBACK_TYPE_PREFIX}.{name}]"


def runnable_feedback_output_selector(name: str) -> str:
    return f"{runnable_feedback_selector(name)}.payload.output"


def runnable_feedback_runnable_ref_selector(name: str) -> str:
    return f"{runnable_feedback_selector(name)}.runnable_ref"



import contextlib
import logging
from collections.abc import Generator

import sqlparse


def param_slot(param_name: str, param_type: str) -> str:
    """Helper function to create a parameter slot for a clickhouse query."""
    return f"{{{param_name}:{param_type}}}"


def safely_format_sql(
    sql: str,
    logger: logging.Logger,
) -> str:
    """Safely format a SQL string with parameters."""
    try:
        return sqlparse.format(sql, reindent=True)
    except:
        logger.info(f"Failed to format SQL: {sql}")
        return sql


# Context for tracking if we're in a NOT operation
class NotContext:
    # Nesting depth
    _depth = 0

    @classmethod
    @contextlib.contextmanager
    def not_context(cls) -> Generator[None, None, None]:
        """Context manager for NOT operations.

        Properly handles nested NOT operations by tracking depth.
        In boolean logic:
        - NOT(expr) flips the result
        - NOT(NOT(expr)) is equivalent to expr
        - NOT(NOT(NOT(expr))) is equivalent to NOT(expr)

        So we only apply special handling when nesting depth is odd.
        """
        cls._depth += 1
        try:
            yield
        finally:
            cls._depth -= 1

    @classmethod
    def is_in_not_context(cls) -> bool:
        """Check if we're in a NOT context with odd nesting depth."""
        return cls._depth % 2 == 1



from typing import Any, Literal

from pydantic import Field, PrivateAttr

import weave
from weave.flow.scorer import WeaveScorerResult
from weave.integrations.bedrock import patch_client


class BedrockGuardrailScorer(weave.Scorer):
    """
    The `BedrockGuardrailScorer` class is a guardrail that leverages AWS Bedrock's
    guardrail functionality to detect and filter content based on configured policies.

    Attributes:
        guardrail_id (str): The identifier of the guardrail to use.
        guardrail_version (str): The version of the guardrail to use.
        source (str): The source of the content to evaluate, either 'INPUT' or 'OUTPUT'.

    Note:
        This scorer requires AWS Bedrock client to be properly configured with
        appropriate permissions to access the guardrail API.
    """

    guardrail_id: str = Field(description="The identifier of the guardrail to use.")
    guardrail_version: str = Field(
        default="DRAFT", description="The version of the guardrail to use."
    )
    source: Literal["INPUT", "OUTPUT"] = Field(
        default="OUTPUT",
        description="The source of the content to evaluate, either 'INPUT' or 'OUTPUT'.",
    )
    bedrock_runtime_kwargs: dict[str, Any] = Field(
        default_factory=dict,
        description="Additional keyword arguments to pass to the Bedrock runtime client.",
    )

    # Private attributes
    _bedrock_runtime: Any = PrivateAttr(default=None)

    def model_post_init(self, __context: Any) -> None:
        """Initialize the Bedrock runtime client."""
        try:
            import boto3

            self._bedrock_runtime = boto3.client(
                "bedrock-runtime", **self.bedrock_runtime_kwargs
            )
            patch_client(self._bedrock_runtime)
        except ImportError:
            raise ImportError(
                "boto3 is not installed. Please install it with 'pip install boto3' "
                "to use the BedrockGuardrailScorer."
            )
        except Exception as e:
            raise RuntimeError(f"Failed to initialize Bedrock runtime client: {e}")

    def format_content(self, output: str) -> dict[str, Any]:
        """Format the content for the guardrail API."""
        return {"source": self.source, "content": [{"text": {"text": output}}]}

    @weave.op
    def score(self, output: str) -> WeaveScorerResult:
        if self._bedrock_runtime is None:
            raise ValueError("Bedrock runtime client is not initialized")

        try:
            # Format the content for the guardrail API
            content = self.format_content(output)

            # Call the ApplyGuardrail API
            response = self._bedrock_runtime.apply_guardrail(
                guardrailIdentifier=self.guardrail_id,
                guardrailVersion=self.guardrail_version,
                source=self.source,
                content=content["content"],
            )

            # Check if the guardrail intervened
            passed = response.get("action") != "GUARDRAIL_INTERVENED"

            assessments = response.get("assessments", [{}])[0]

            # Get the modified output if available
            modified_output = None
            if response.get("outputs") and len(response["outputs"]) > 0:
                modified_output = response["outputs"][0].get("text")

            return WeaveScorerResult(
                passed=passed,
                metadata={
                    "modified_output": modified_output,
                    "usage": response.get("usage", {}),
                    "assessments": assessments,
                },
            )
        except Exception as e:
            return WeaveScorerResult(
                passed=False,
                metadata={
                    "error": f"Error applying Bedrock guardrail: {str(e)}",
                    "reason": f"Error applying Bedrock guardrail: {str(e)}",
                },
            )



from typing import TYPE_CHECKING, Any, Union

import pydantic

if TYPE_CHECKING:
    from pydantic.v1 import BaseModel as PydanticBaseModelV1


def is_pydantic_v1_base_model(obj: Any) -> bool:
    try:
        from pydantic.v1 import BaseModel as PydanticBaseModelV1

        return isinstance(obj, PydanticBaseModelV1)
    except ImportError:
        return False


PydanticBaseModelGeneral = Union[pydantic.BaseModel, "PydanticBaseModelV1"]


def pydantic_model_fields(
    obj: PydanticBaseModelGeneral,
) -> dict[str, pydantic.fields.FieldInfo]:
    if isinstance(obj, pydantic.BaseModel):
        return obj.model_fields
    elif is_pydantic_v1_base_model(obj):
        return obj.__fields__
    else:
        raise TypeError(f"{obj} is not a pydantic model")


def pydantic_asdict_one_level(obj: PydanticBaseModelGeneral) -> dict[str, Any]:
    """
    This is equivalent to `obj.model_dump(by_alias=True)`, but does not recursively
    convert nested pydantic objects to dicts. This is particularly useful when you want
    manually iterate over the fields of a pydantic object and do something with them.
    """
    fields = pydantic_model_fields(obj)
    final = {}
    for prop_name, field in fields.items():
        use_name = prop_name
        # This odd check is to support different pydantic versions
        if hasattr(field, "exclude") and field.exclude:
            continue
        # This odd check is to support different pydantic versions
        if hasattr(field, "alias") and field.alias:
            use_name = field.alias
        final[use_name] = getattr(obj, prop_name)
    return final



from typing import Any

from pydantic import BaseModel, ValidationError

import weave


class PydanticScorer(weave.Scorer):
    """Validate the model output against a pydantic model."""

    model: type[BaseModel]

    @weave.op
    def score(self, output: Any) -> dict:
        if isinstance(output, str):
            try:
                self.model.model_validate_json(output)
            except ValidationError:
                return {"valid_pydantic": False}
            else:
                return {"valid_pydantic": True}
        else:
            try:
                self.model.model_validate(output)
            except ValidationError:
                return {"valid_pydantic": False}
            else:
                return {"valid_pydantic": True}



"""
Query optimization framework for call queries.

Optimizes complex queries on call data before expensive database aggregation (groupby).
Optimizations should be conservative, not guaranteed to full filter down but should
   be a superset of results.

Key strategies:
1. String optimization - Uses LIKE patterns for string fields
2. ID-based optimization - Uses sortable_datetime column for datetime filtering

Optimization SQL is applied before GROUP BY, reducing memory usage and
improving performance for complex conditions.
"""

import datetime
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Optional, Union

from pydantic import BaseModel

from weave.trace_server.calls_query_builder.utils import (
    NotContext,
    param_slot,
)
from weave.trace_server.interface import query as tsi_query

if TYPE_CHECKING:
    from weave.trace_server.calls_query_builder.calls_query_builder import (
        Condition,
        ParamBuilder,
    )

START_ONLY_CALL_FIELDS = {"started_at", "inputs_dump", "attributes_dump"}
END_ONLY_CALL_FIELDS = {"ended_at", "output_dump", "summary_dump"}
STRING_FIELDS_TO_OPTIMIZE = {"inputs_dump", "output_dump", "attributes_dump"}
DATETIME_FIELDS_TO_OPTIMIZE = {"started_at"}

DATETIME_BUFFER_TIME_SECONDS = 60 * 5  # 5 minutes


def _field_requires_null_check(field: str) -> bool:
    """Returns whether a string is a start or end field

    Start and end fields require appending a NULL to all conditional
    checks when operating on the calls_merged table, as unmerged
    call starts and ends can be valid results without inputs/output
    """
    return field in START_ONLY_CALL_FIELDS | END_ONLY_CALL_FIELDS


def _can_optimize_string_field(field: str) -> bool:
    """Returns whether a field can be optimized for string search."""
    return field in STRING_FIELDS_TO_OPTIMIZE


def _can_optimize_datetime_field(field: str) -> bool:
    """Returns whether a field can be optimized for datetime filtering.

    Currently this is only the started_at field, which can be converted
    to the `sortable_datetime` column.
    """
    return field in DATETIME_FIELDS_TO_OPTIMIZE


class QueryOptimizationProcessor(ABC):
    """
    Abstract base class for query optimization processors.

    This class defines the interface for optimization processors that convert
    query operations into optimized SQL conditions. Subclasses should implement
    specific strategies for different types of data or operations.
    """

    def __init__(self, pb: "ParamBuilder", table_alias: str):
        self.pb = pb
        self.table_alias = table_alias

    def process_operand(self, operand: tsi_query.Operand) -> Optional[str]:
        """
        Process an operand and convert it to an optimized SQL condition if possible.

        Ignores leaf operations like Literal, GetField, and Convert.

        Args:
            operand: The operand to process

        Returns:
            SQL condition string or None if optimization is not possible
        """
        if isinstance(operand, tsi_query.LiteralOperation):
            return None
        elif isinstance(operand, tsi_query.GetFieldOperator):
            return None
        elif isinstance(operand, tsi_query.ConvertOperation):
            return None
        return apply_processor(self, operand)

    def process_and(self, operation: tsi_query.AndOperation) -> Optional[str]:
        """Process AND operations into optimized SQL."""
        conditions = []
        for op in operation.and_:
            result = self.process_operand(op)
            if result:
                conditions.append(result)

        if conditions:
            return "(" + " AND ".join(conditions) + ")"
        return None

    def process_or(self, operation: tsi_query.OrOperation) -> Optional[str]:
        """
        Process OR operations into optimized SQL.

        Note: If any condition in an OR cannot be optimized, the entire OR
        cannot be optimized, so we return None.
        """
        conditions = []
        for op in operation.or_:
            result = self.process_operand(op)
            if result is None:
                # If any or condition can't be optimized, return
                # TODO: this should return the non optimized,
                # non aggreagated condition when available
                return None
            conditions.append(result)

        if conditions:
            return "(" + " OR ".join(conditions) + ")"
        return None

    def process_not(self, operation: tsi_query.NotOperation) -> Optional[str]:
        """Process NOT operations into optimized SQL."""
        if len(operation.not_) != 1:
            return None

        with NotContext.not_context():
            result = self.process_operand(operation.not_[0])

        if result is None:
            return None
        return f"NOT ({result})"

    @abstractmethod
    def process_eq(self, operation: tsi_query.EqOperation) -> Optional[str]:
        """Process equality operation."""
        pass

    @abstractmethod
    def process_contains(self, operation: tsi_query.ContainsOperation) -> Optional[str]:
        """Process contains operation."""
        pass

    @abstractmethod
    def process_in(self, operation: tsi_query.InOperation) -> Optional[str]:
        """Process in operation."""
        pass

    @abstractmethod
    def process_gt(self, operation: tsi_query.GtOperation) -> Optional[str]:
        """Process greater than operation."""
        pass

    @abstractmethod
    def process_gte(self, operation: tsi_query.GteOperation) -> Optional[str]:
        """Process greater than or equal operation."""
        pass

    def finalize_sql(self, result: Optional[str]) -> Optional[str]:
        """
        Final step to make valid SQL for the calls query.

        This method can be overridden by subclasses to apply additional
        transformations to the final SQL condition.

        Args:
            result: The SQL condition to finalize

        Returns:
            The finalized SQL condition or None if no condition
        """
        if result:
            return f"AND {result}"
        return None


class StringOptimizationProcessor(QueryOptimizationProcessor):
    """
    Optimization processor for string operations.

    This processor creates LIKE-based SQL conditions to optimize queries
    on string fields before aggregation, reducing memory pressure.
    """

    def process_eq(self, operation: tsi_query.EqOperation) -> Optional[str]:
        """
        Process equality operation on string fields.

        Creates SQL condition using LIKE patterns for strings in JSON fields.
        """
        return _create_like_optimized_eq_condition(operation, self.pb, self.table_alias)

    def process_contains(self, operation: tsi_query.ContainsOperation) -> Optional[str]:
        """
        Process contains operation on string fields.

        Creates SQL condition using LIKE patterns for substrings in JSON fields.
        """
        return _create_like_optimized_contains_condition(
            operation, self.pb, self.table_alias
        )

    def process_in(self, operation: tsi_query.InOperation) -> Optional[str]:
        """
        Process IN operation on string fields.

        Creates SQL conditions using LIKE patterns for multiple string values.
        """
        return _create_like_optimized_in_condition(operation, self.pb, self.table_alias)

    def process_gt(self, operation: tsi_query.GtOperation) -> Optional[str]:
        """Not implemented for string optimization."""
        return None

    def process_gte(self, operation: tsi_query.GteOperation) -> Optional[str]:
        """Not implemented for string optimization."""
        return None


class SortableDatetimeOptimizationProcessor(QueryOptimizationProcessor):
    """
    Optimization processor for sortable_datetime operations.

    This processor creates SQL conditions that filter based on the
    `sortable_datetime` column, which can significantly reduce the dataset before
    doing more complex operations after aggregation.
    """

    def process_eq(self, operation: tsi_query.EqOperation) -> Optional[str]:
        """Not implemented for sortable_datetime optimization."""
        return None

    def process_contains(self, operation: tsi_query.ContainsOperation) -> Optional[str]:
        """Not implemented for sortable_datetime optimization."""
        return None

    def process_in(self, operation: tsi_query.InOperation) -> Optional[str]:
        """Not implemented for sortable_datetime optimization."""
        return None

    def process_gt(self, operation: tsi_query.GtOperation) -> Optional[str]:
        """
        Process GT operation on sortable_datetime fields using sortable_datetime optimization.

        Creates SQL condition that filters started_at with the sortable_datetime column.
        """
        return _create_datetime_optimization_sql(
            operation, self.pb, self.table_alias, ">"
        )

    def process_gte(self, operation: tsi_query.GteOperation) -> Optional[str]:
        """
        Process GTE operation on sortable_datetime fields using sortable_datetime optimization.

        Creates SQL condition that filters started_at with the sortable_datetime column.
        """
        return _create_datetime_optimization_sql(
            operation, self.pb, self.table_alias, ">="
        )


def apply_processor(
    processor: QueryOptimizationProcessor, operation: tsi_query.Operation
) -> Optional[str]:
    if isinstance(operation, tsi_query.AndOperation):
        return processor.process_and(operation)
    elif isinstance(operation, tsi_query.OrOperation):
        return processor.process_or(operation)
    elif isinstance(operation, tsi_query.NotOperation):
        return processor.process_not(operation)
    elif isinstance(operation, tsi_query.EqOperation):
        return processor.process_eq(operation)
    elif isinstance(operation, tsi_query.ContainsOperation):
        return processor.process_contains(operation)
    elif isinstance(operation, tsi_query.InOperation):
        return processor.process_in(operation)
    elif isinstance(operation, tsi_query.GtOperation):
        return processor.process_gt(operation)
    elif isinstance(operation, tsi_query.GteOperation):
        return processor.process_gte(operation)
    return None


class OptimizationConditions(BaseModel):
    str_filter_opt_sql: Optional[str] = None
    sortable_datetime_filters_sql: Optional[str] = None


def process_query_to_optimization_sql(
    conditions: list["Condition"],
    param_builder: "ParamBuilder",
    table_alias: str,
) -> OptimizationConditions:
    """Converts a list of conditions to optimization conditions for a clickhouse query.

    This function creates SQL conditions that can be applied before the GROUP BY
    to filter out rows that definitely won't match the heavy conditions. These
    conditions MUST be identical or less restrictive than the conditions in the
    `conditions` list which will appear in HAVING after group by.

    For fields that may only exist in start or end parts, we add special handling
    to avoid filtering out rows where the field is NULL (as they might be part of
    a valid call when combined with other parts).

    Performance note: This optimization is critical for queries with heavy fields,
    as it can significantly reduce peak memory by filtering before aggregation.
    """
    if not conditions:
        return OptimizationConditions()

    # Create a single AND operation from all conditions
    and_operation = tsi_query.AndOperation(**{"$and": [c.operand for c in conditions]})

    # Apply string optimization
    string_processor = StringOptimizationProcessor(param_builder, table_alias)
    string_result = apply_processor(string_processor, and_operation)
    string_result_sql = string_processor.finalize_sql(string_result)

    # Apply sortable_datetime optimization
    sortable_datetime_processor = SortableDatetimeOptimizationProcessor(
        param_builder, table_alias
    )
    sortable_datetime_result = apply_processor(
        sortable_datetime_processor, and_operation
    )
    sortable_datetime_result_sql = sortable_datetime_processor.finalize_sql(
        sortable_datetime_result
    )

    return OptimizationConditions(
        str_filter_opt_sql=string_result_sql,
        sortable_datetime_filters_sql=sortable_datetime_result_sql,
    )


def _create_like_condition(
    field: str,
    like_pattern: str,
    pb: "ParamBuilder",
    table_alias: str,
    case_insensitive: bool = False,
) -> str:
    """Creates a LIKE condition for a JSON field."""
    field_name = f"{table_alias}.{field}"

    if case_insensitive:
        param_name = pb.add_param(like_pattern.lower())
        return f"lower({field_name}) LIKE {param_slot(param_name, 'String')}"
    else:
        param_name = pb.add_param(like_pattern)
        return f"{field_name} LIKE {param_slot(param_name, 'String')}"


def _extract_field_and_literal(
    operation: Union[
        tsi_query.EqOperation, tsi_query.GtOperation, tsi_query.GteOperation
    ],
) -> tuple[Optional[tsi_query.GetFieldOperator], Optional[tsi_query.LiteralOperation]]:
    """Extract field and literal operands from a binary operation.

    Returns a tuple of (field_operand, literal_operand) or (None, None) if invalid.
    """
    ops = (
        operation.eq_
        if hasattr(operation, "eq_")
        else (operation.gt_ if hasattr(operation, "gt_") else operation.gte_)
    )

    if len(ops) != 2:
        return None, None

    field_operand = None
    literal_operand = None

    if isinstance(ops[0], tsi_query.GetFieldOperator):
        field_operand = ops[0]
        literal_operand = ops[1]
    elif isinstance(ops[1], tsi_query.GetFieldOperator):
        field_operand = ops[1]
        literal_operand = ops[0]

    if not isinstance(literal_operand, tsi_query.LiteralOperation):
        return None, None

    return field_operand, literal_operand


def _create_like_optimized_eq_condition(
    operation: tsi_query.EqOperation,
    pb: "ParamBuilder",
    table_alias: str,
) -> Optional[str]:
    """Creates a LIKE-optimized condition for equality operations."""
    field_operand, literal_operand = _extract_field_and_literal(operation)
    if field_operand is None or literal_operand is None:
        return None

    # Return if literal isn't a string
    if not isinstance(literal_operand.literal_, str):
        return None

    from weave.trace_server.calls_query_builder.calls_query_builder import (
        get_field_by_name,
    )

    field = get_field_by_name(field_operand.get_field_).field
    literal_value = literal_operand.literal_

    if not _can_optimize_string_field(field):
        return None

    if not literal_value:
        # Empty string is not a valid value for LIKE optimization
        return None

    # Boolean literals are not wrapped in quotes in JSON payloads
    if literal_value in ("true", "false"):
        like_pattern = f"%{literal_value}%"
    else:
        like_pattern = f'%"{literal_value}"%'

    like_condition = _create_like_condition(field, like_pattern, pb, table_alias)
    if _field_requires_null_check(field):
        return f"({like_condition} OR {table_alias}.{field} IS NULL)"
    return like_condition


def _create_like_optimized_contains_condition(
    operation: tsi_query.ContainsOperation,
    pb: "ParamBuilder",
    table_alias: str,
) -> Optional[str]:
    """Creates a LIKE-optimized condition for contains operations."""
    # Check if the input is a GetField operation on a JSON field
    if not isinstance(operation.contains_.input, tsi_query.GetFieldOperator):
        return None
    # Return if substr isn't a string literal
    if not isinstance(
        operation.contains_.substr, tsi_query.LiteralOperation
    ) or not isinstance(operation.contains_.substr.literal_, str):
        return None

    from weave.trace_server.calls_query_builder.calls_query_builder import (
        get_field_by_name,
    )

    field = get_field_by_name(operation.contains_.input.get_field_).field
    substr_value = operation.contains_.substr.literal_
    if not substr_value:
        # Empty string is not a valid value for LIKE optimization
        return None

    if not _can_optimize_string_field(field):
        return None

    case_insensitive = operation.contains_.case_insensitive or False
    like_pattern = f'%"%{substr_value}%"%'

    like_condition = _create_like_condition(
        field, like_pattern, pb, table_alias, case_insensitive
    )
    if _field_requires_null_check(field):
        return f"({like_condition} OR {table_alias}.{field} IS NULL)"
    return like_condition


def _create_like_optimized_in_condition(
    operation: tsi_query.InOperation,
    pb: "ParamBuilder",
    table_alias: str,
) -> Optional[str]:
    """Creates a LIKE-optimized condition for in operations."""
    # Check if the left side is a GetField operation on a JSON field
    if not isinstance(operation.in_[0], tsi_query.GetFieldOperator):
        return None
    # Return if right-side isn't non-empty list
    if (
        len(operation.in_) != 2
        or not isinstance(operation.in_[1], list)
        or len(operation.in_[1]) == 0
    ):
        return None

    from weave.trace_server.calls_query_builder.calls_query_builder import (
        get_field_by_name,
    )

    field = get_field_by_name(operation.in_[0].get_field_).field
    if not _can_optimize_string_field(field):
        return None

    # Create OR conditions for each value
    like_conditions: list[str] = []

    for value_operand in operation.in_[1]:
        if (
            not isinstance(value_operand, tsi_query.LiteralOperation)
            or not isinstance(value_operand.literal_, str)
            or not value_operand.literal_
        ):
            return None

        like_pattern = f'%"{value_operand.literal_}"%'
        like_condition = _create_like_condition(field, like_pattern, pb, table_alias)
        like_conditions.append(like_condition)

    or_sql = "(" + " OR ".join(like_conditions) + ")"
    if _field_requires_null_check(field):
        return f"({or_sql} OR {table_alias}.{field} IS NULL)"
    return or_sql


def _timestamp_to_datetime_str(timestamp: int) -> str:
    """Converts a timestamp to a datetime string."""
    return datetime.datetime.fromtimestamp(
        timestamp, tz=datetime.timezone.utc
    ).strftime("%Y-%m-%d %H:%M:%S.%f")


def _create_datetime_optimization_sql(
    operation: Union[tsi_query.GtOperation, tsi_query.GteOperation],
    pb: "ParamBuilder",
    table_alias: str,
    op_str: str,
) -> Optional[str]:
    """Creates SQL for datetime optimization using indexed sortable_datetime column.

    Applies a buffer to the timestamp to make the filter more permissive:
    - For normal context: Subtracts buffer from timestamp
    - For NOT context: Adds buffer to timestamp
    """
    field_operand, literal_operand = _extract_field_and_literal(operation)
    if field_operand is None or literal_operand is None:
        return None

    field_name = field_operand.get_field_
    if not _can_optimize_datetime_field(field_name):
        return None

    literal_value = literal_operand.literal_

    if not literal_value or not isinstance(literal_value, (int, float)):
        return None

    # convert timestamp to datetime_str
    timestamp = int(literal_value)

    # Apply buffer in appropriate direction based on context
    buffer_seconds = int(DATETIME_BUFFER_TIME_SECONDS)
    if NotContext.is_in_not_context():
        # For NOT context, add buffer to make filter more permissive
        timestamp += buffer_seconds
    else:
        # For normal context, subtract buffer to make filter more permissive
        timestamp -= buffer_seconds

    datetime_str = _timestamp_to_datetime_str(timestamp)

    param_name = pb.add_param(datetime_str)
    return (
        f"{table_alias}.sortable_datetime {op_str} {param_slot(param_name, 'String')}"
    )



import importlib
from typing import TYPE_CHECKING, Any, Callable, Union

from pydantic import BaseModel

import weave
from weave.integrations.patcher import SymbolPatcher
from weave.trace.autopatch import OpSettings
from weave.trace.op import Op
from weave.trace.serialization.serialize import dictify

if TYPE_CHECKING:
    from dspy.primitives.prediction import Example


def get_symbol_patcher(
    base_symbol: str, attribute_name: str, settings: OpSettings
) -> SymbolPatcher:
    display_name = base_symbol + "." + attribute_name
    display_name = (
        display_name.replace(".__call__", "")
        if attribute_name.endswith(".__call__")
        else display_name
    )
    return SymbolPatcher(
        lambda: importlib.import_module(base_symbol),
        attribute_name,
        dspy_wrapper(
            settings.model_copy(update={"name": settings.name or display_name})
        ),
    )


def dump_dspy_objects(data: Any) -> Any:
    import numpy as np
    from dspy import Example, Module

    if isinstance(data, Example):
        return data.toDict()

    elif isinstance(data, Module):
        return data.dump_state()

    elif isinstance(data, np.ndarray):
        return data.tolist()

    elif isinstance(data, dict):
        return {key: dump_dspy_objects(value) for key, value in data.items()}

    elif isinstance(data, list):
        return [dump_dspy_objects(item) for item in data]

    return data


def dspy_postprocess_inputs(inputs: dict[str, Any]) -> dict[str, Any]:
    from dspy import Adapter, Evaluate, Predict

    if "self" in inputs:
        dictified_inputs_self = dictify(inputs["self"])
        if dictified_inputs_self["__class__"]["module"] == "__main__":
            dictified_inputs_self["__class__"]["module"] = ""

        # Serialize the signature of the object if it is a Predict or Adapter
        if isinstance(inputs["self"], Predict) or isinstance(inputs["self"], Adapter):
            if hasattr(inputs["self"], "signature"):
                if isinstance(inputs["self"].signature, BaseModel):
                    dictified_inputs_self["signature"] = inputs[
                        "self"
                    ].signature.model_json_schema()
                else:
                    dictified_inputs_self["signature"] = inputs["self"].signature

        dictified_inputs_self = dump_dspy_objects(dictified_inputs_self)

        # Recursively serialize the dspy objects in the devset
        if isinstance(inputs["self"], Evaluate):
            dictified_inputs_self["devset"] = [
                dump_dspy_objects(example) for example in inputs["self"].devset
            ]

            # Convert the metric to a weave op if it is not already one
            if hasattr(inputs["self"], "metric"):
                inputs["self"].metric = (
                    weave.op(inputs["self"].metric)
                    if not isinstance(inputs["self"].metric, Op)
                    else inputs["self"].metric
                )

        inputs["self"] = dictified_inputs_self

    return dump_dspy_objects(inputs)


def dspy_postprocess_outputs(
    outputs: Union[Any, "Example"],
) -> Union[list[Any], dict[str, Any], Any]:
    import numpy as np
    from dspy import Example, Module

    if isinstance(outputs, Module):
        outputs = outputs.dump_state()

    if isinstance(outputs, Example):
        outputs = outputs.toDict()

    if isinstance(outputs, np.ndarray):
        outputs = outputs.tolist()

    return dump_dspy_objects(outputs)


def dspy_wrapper(settings: OpSettings) -> Callable[[Callable], Callable]:
    def wrapper(fn: Callable) -> Callable:
        op_kwargs = settings.model_dump()
        if not op_kwargs.get("postprocess_inputs"):
            op_kwargs["postprocess_inputs"] = dspy_postprocess_inputs
        if not op_kwargs.get("postprocess_output"):
            op_kwargs["postprocess_output"] = dspy_postprocess_outputs
        op = weave.op(fn, **op_kwargs)
        return op

    return wrapper


def get_op_name_for_callback(instance: Any, inputs: dict[str, Any]) -> str:
    instance_class_name = instance.__class__.__name__
    return (
        f"dspy.{instance_class_name}"
        if "dspy." in inputs["self"]["__class__"]["module"]
        else instance_class_name
    )



"""Default model configurations for different LLM providers."""

OPENAI_DEFAULT_MODEL = "gpt-4o"
OPENAI_DEFAULT_EMBEDDING_MODEL = "text-embedding-3-small"
OPENAI_DEFAULT_MODERATION_MODEL = "omni-moderation-latest"

ANTHROPIC_DEFAULT_MODEL = "claude-3-5-sonnet"

MISTRAL_DEFAULT_MODEL = "mistral-large-latest"
MISTRAL_DEFAULT_EMBEDDING_MODEL = "mistral-embed"

DEFAULT_MAX_TOKENS = 4096
DEFAULT_TEMPERATURE = 0.7

# Model paths for various scorers
MODEL_PATHS = {
    "toxicity_scorer": "c-metrics/weave-scorers/toxicity_scorer:v0",
    "bias_scorer": "c-metrics/weave-scorers/bias_scorer:v0",
    "coherence_scorer": "c-metrics/weave-scorers/coherence_scorer:v0",
    "fluency_scorer": "c-metrics/weave-scorers/fluency_scorer:v0",
    "hallucination_scorer": "c-metrics/weave-scorers/hallucination_hhem_scorer:v0",
    "relevance_scorer": "c-metrics/weave-scorers/relevance_scorer:v0",
}



from functools import wraps
from typing import Any, Callable, Optional

from pydantic import BaseModel

import weave
from weave.trace.autopatch import OpSettings
from weave.trace.op import _add_accumulator


def instructor_iterable_accumulator(
    acc: Optional[list[BaseModel]], value: BaseModel
) -> list[BaseModel]:
    if acc is None:
        return [value]
    if acc[-1] != value:
        acc.append(value)
    return acc


def should_accumulate_iterable(inputs: dict) -> bool:
    if isinstance(inputs, dict):
        if "stream" in inputs:
            return inputs["stream"]
        elif "kwargs" in inputs:
            if "stream" in inputs["kwargs"]:
                return inputs.get("kwargs", {}).get("stream")
    return False


def instructor_wrapper_sync(settings: OpSettings) -> Callable[[Callable], Callable]:
    def wrapper(fn: Callable) -> Callable:
        op_kwargs = settings.model_dump()
        op = weave.op(fn, **op_kwargs)
        return _add_accumulator(
            op,  # type: ignore
            make_accumulator=lambda inputs: instructor_iterable_accumulator,
            should_accumulate=should_accumulate_iterable,
        )

    return wrapper


def instructor_wrapper_async(settings: OpSettings) -> Callable[[Callable], Callable]:
    def wrapper(fn: Callable) -> Callable:
        def _fn_wrapper(fn: Callable) -> Callable:
            @wraps(fn)
            async def _async_wrapper(*args: Any, **kwargs: Any) -> Any:
                return await fn(*args, **kwargs)

            return _async_wrapper

        op_kwargs = settings.model_dump()
        op = weave.op(_fn_wrapper(fn), **op_kwargs)
        return _add_accumulator(
            op,  # type: ignore
            make_accumulator=lambda inputs: instructor_iterable_accumulator,
            should_accumulate=should_accumulate_iterable,
        )

    return wrapper



"""Human-readable wrapper classes for OpenTelemetry trace protocol buffer definitions.

This module provides simple, human-readable Python classes that represent the
trace protocol buffer definitions from opentelemetry.proto.trace.v1.trace_pb2.
"""

import datetime
from binascii import hexlify
from collections.abc import Iterator
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Optional

from opentelemetry.proto.common.v1.common_pb2 import InstrumentationScope
from opentelemetry.proto.resource.v1.resource_pb2 import Resource as PbResource
from opentelemetry.proto.trace.v1.trace_pb2 import (
    ResourceSpans as PbResourceSpans,
)
from opentelemetry.proto.trace.v1.trace_pb2 import (
    ScopeSpans as PbScopeSpans,
)
from opentelemetry.proto.trace.v1.trace_pb2 import (
    Span as PbSpan,
)
from opentelemetry.proto.trace.v1.trace_pb2 import (
    Status as PbStatus,
)
from opentelemetry.proto.trace.v1.trace_pb2 import (
    TracesData as PbTracesData,
)

from weave.trace_server import trace_server_interface as tsi

from .attributes import (
    Attributes,
    AttributesFactory,
    to_json_serializable,
    unflatten_key_values,
)


class SpanKind(Enum):
    """Enum representing the span's kind."""

    UNSPECIFIED = 0
    INTERNAL = 1
    SERVER = 2
    CLIENT = 3
    PRODUCER = 4
    CONSUMER = 5

    @classmethod
    def from_proto(cls, proto_kind: int) -> "SpanKind":
        return cls(proto_kind)


class StatusCode(Enum):
    """Enum representing the span's status code."""

    UNSET = 0
    OK = 1
    ERROR = 2

    @classmethod
    def from_proto(cls, proto_code: int) -> "StatusCode":
        """Convert from protobuf enum value to StatusCode."""
        return cls(proto_code)


@dataclass
class Status:
    """Represents the status of a span."""

    code: StatusCode = StatusCode.UNSET
    message: str = ""

    @classmethod
    def from_proto(cls, proto_status: PbStatus) -> "Status":
        """Create a Status from a protobuf Status."""
        return cls(
            code=StatusCode.from_proto(proto_status.code), message=proto_status.message
        )

    def as_weave_status(self) -> Optional[tsi.TraceStatus]:
        """Convert from protobuf enum value to StatusCode."""
        if self.code == StatusCode.OK:
            return tsi.TraceStatus.SUCCESS
        elif self.code == StatusCode.ERROR:
            return tsi.TraceStatus.ERROR
        # UNSET: This is not 'running' because if the trace was sent the call completed
        return None

    def as_dict(self) -> dict[str, Any]:
        return to_json_serializable(
            {
                "code": self.code.name,
                "message": self.message,
            }
        )


@dataclass
class Event:
    """Represents a timed event in a span."""

    name: str
    timestamp: int  # nanoseconds since epoch
    attributes: dict[str, Any] = field(default_factory=dict)
    dropped_attributes_count: int = 0

    @property
    def datetime(self) -> datetime.datetime:
        """Return the timestamp as a datetime object."""
        return datetime.datetime.fromtimestamp(self.timestamp / 1_000_000_000)

    @classmethod
    def from_proto(cls, proto_event: PbSpan.Event) -> "Event":
        """Create an Event from a protobuf Event."""
        return cls(
            name=proto_event.name,
            timestamp=proto_event.time_unix_nano,
            attributes=unflatten_key_values(proto_event.attributes),
            dropped_attributes_count=proto_event.dropped_attributes_count,
        )


@dataclass
class Link:
    """Represents a link to another span."""

    trace_id: str
    span_id: str
    trace_state: str = ""
    attributes: dict[str, Any] = field(default_factory=dict)
    dropped_attributes_count: int = 0
    flags: int = 0

    @classmethod
    def from_proto(cls, proto_link: PbSpan.Link) -> "Link":
        """Create a Link from a protobuf Link."""
        return cls(
            trace_id=hexlify(proto_link.trace_id).decode("ascii"),
            span_id=hexlify(proto_link.span_id).decode("ascii"),
            trace_state=proto_link.trace_state,
            attributes=unflatten_key_values(proto_link.attributes),
            dropped_attributes_count=proto_link.dropped_attributes_count,
            flags=proto_link.flags,
        )


@dataclass
class Resource:
    attributes: dict[str, Any] = field(default_factory=dict)
    dropped_attributes_count: int = 0

    @classmethod
    def from_proto(cls, proto_resource: PbResource) -> "Resource":
        attributes = {}
        if proto_resource.attributes:
            attributes = unflatten_key_values(proto_resource.attributes)
        dropped_attributes_count = 0
        if proto_resource.dropped_attributes_count:
            dropped_attributes_count = proto_resource.dropped_attributes_count
        return cls(attributes, dropped_attributes_count)

    def as_dict(self) -> dict[str, Any]:
        return {
            "attributes": self.attributes,
            "dropped_attributes_count": self.dropped_attributes_count,
        }


@dataclass
class Span:
    """Represents a span in a trace."""

    resource: Optional[Resource]
    name: str
    trace_id: str
    span_id: str
    start_time_unix_nano: int
    end_time_unix_nano: int
    attributes: Attributes
    kind: SpanKind = SpanKind.UNSPECIFIED
    parent_id: Optional[str] = None
    trace_state: str = ""
    flags: int = 0
    dropped_attributes_count: int = 0
    events: list[Event] = field(default_factory=list)
    dropped_events_count: int = 0
    links: list[Link] = field(default_factory=list)
    dropped_links_count: int = 0
    status: Status = field(default_factory=Status)

    @property
    def start_time(self) -> datetime.datetime:
        """Return the start time as a datetime object."""
        return datetime.datetime.fromtimestamp(
            self.start_time_unix_nano / 1_000_000_000
        )

    @property
    def end_time(self) -> datetime.datetime:
        """Return the end time as a datetime object."""
        return datetime.datetime.fromtimestamp(self.end_time_unix_nano / 1_000_000_000)

    @property
    def duration_ns(self) -> int:
        """Return the duration in nanoseconds."""
        return self.end_time_unix_nano - self.start_time_unix_nano

    @property
    def duration_ms(self) -> float:
        """Return the duration in milliseconds."""
        return self.duration_ns / 1_000_000

    @classmethod
    def from_proto(
        cls, proto_span: PbSpan, resource: Optional[Resource] = None
    ) -> "Span":
        """Create a Span from a protobuf Span."""
        parent_id = None
        if proto_span.parent_span_id:
            parent_id = hexlify(proto_span.parent_span_id).decode("ascii")

        return cls(
            name=proto_span.name,
            trace_id=hexlify(proto_span.trace_id).decode("ascii"),
            span_id=hexlify(proto_span.span_id).decode("ascii"),
            start_time_unix_nano=proto_span.start_time_unix_nano,
            end_time_unix_nano=proto_span.end_time_unix_nano,
            kind=SpanKind.from_proto(proto_span.kind),
            parent_id=parent_id,
            trace_state=proto_span.trace_state,
            flags=proto_span.flags,
            attributes=AttributesFactory().from_proto(key_values=proto_span.attributes),
            dropped_attributes_count=proto_span.dropped_attributes_count,
            events=[Event.from_proto(e) for e in proto_span.events],
            dropped_events_count=proto_span.dropped_events_count,
            links=[Link.from_proto(l) for l in proto_span.links],
            dropped_links_count=proto_span.dropped_links_count,
            status=Status.from_proto(proto_span.status),
            resource=resource,
        )

    # The full OTEL Span as it is recieved
    def as_dict(self) -> dict[str, Any]:
        return to_json_serializable(
            {
                "name": self.name,
                "context": {
                    "trace_id": self.trace_id,
                    "span_id": self.span_id,
                    "trace_state": self.trace_state,
                },
                "kind": self.kind.name,
                "parent_id": self.parent_id,
                "start_time": self.start_time,
                "end_time": self.end_time,
                "status": self.status.as_dict(),
                "attributes": self.attributes._attributes,
                "events": self.events,
                "links": self.links,
                "resource": self.resource.as_dict() if self.resource else None,
            }
        )

    def to_call(
        self, project_id: str
    ) -> tuple[tsi.StartedCallSchemaForInsert, tsi.EndedCallSchemaForInsert]:
        summary_insert_map = self.attributes.get_weave_summary()
        inputs = self.attributes.get_weave_inputs()
        outputs = self.attributes.get_weave_outputs()
        attributes = self.attributes.get_weave_attributes(
            extra={"otel_span": self.as_dict()}
        )

        # Options: set
        start_call = tsi.StartedCallSchemaForInsert(
            project_id=project_id,
            id=self.span_id,
            op_name=self.name,
            trace_id=self.trace_id,
            parent_id=self.parent_id,
            started_at=self.start_time,
            attributes=attributes,
            inputs=inputs,
            wb_user_id=None,
            wb_run_id=None,
        )
        exception_msg = (
            self.status.message if self.status.code == StatusCode.ERROR else None
        )

        end_call = tsi.EndedCallSchemaForInsert(
            project_id=project_id,
            id=self.span_id,
            ended_at=self.end_time,
            exception=exception_msg,
            output=outputs,
            summary=summary_insert_map,
        )
        return (start_call, end_call)


@dataclass
class ScopeSpans:
    """Represents a collection of spans from a specific instrumentation scope."""

    scope: InstrumentationScope
    spans: list[Span] = field(default_factory=list)
    schema_url: str = ""

    def __iter__(self) -> Iterator[Span]:
        yield from self.spans

    @classmethod
    def from_proto(
        cls, proto_scope_spans: PbScopeSpans, resource: Optional[Resource] = None
    ) -> "ScopeSpans":
        """Create a ScopeSpans from a protobuf ScopeSpans."""
        return cls(
            scope=proto_scope_spans.scope,
            spans=[Span.from_proto(s, resource) for s in proto_scope_spans.spans],
            schema_url=proto_scope_spans.schema_url,
        )


@dataclass
class ResourceSpans:
    """Represents a collection of spans from a specific resource."""

    resource: Optional[Resource]
    scope_spans: list[ScopeSpans] = field(default_factory=list)
    schema_url: str = ""

    def __iter__(self) -> Iterator[ScopeSpans]:
        yield from self.scope_spans

    @classmethod
    def from_proto(cls, proto_resource_spans: PbResourceSpans) -> "ResourceSpans":
        """Create a ResourceSpans from a protobuf ResourceSpans."""
        resource = Resource.from_proto(proto_resource_spans.resource)
        return cls(
            resource=resource,
            scope_spans=[
                ScopeSpans.from_proto(s, resource)
                for s in proto_resource_spans.scope_spans
            ],
            schema_url=proto_resource_spans.schema_url,
        )


@dataclass
class TracesData:
    """Top-level collection of trace data."""

    resource_spans: list[ResourceSpans] = field(default_factory=list)

    def __iter__(self) -> Iterator[ResourceSpans]:
        yield from self.resource_spans

    @classmethod
    def from_proto(cls, proto_traces_data: PbTracesData) -> "TracesData":
        """Create a TracesData from a protobuf TracesData."""
        return cls(
            resource_spans=[
                ResourceSpans.from_proto(rs) for rs in proto_traces_data.resource_spans
            ]
        )



from typing import Optional

import pydantic

RefStr = str


# This is just an alternative to weave.Object for the server side.
# I _think_ this will go away once we have the full weave system on the server
class BaseObject(pydantic.BaseModel):
    name: Optional[str] = None
    description: Optional[str] = None



import asyncio
import logging
import traceback
from datetime import datetime
from itertools import chain, repeat
from typing import Any, Callable, Literal, Optional, Union

from pydantic import PrivateAttr
from rich import print
from rich.console import Console
from typing_extensions import Self

import weave
from weave.flow import util
from weave.flow.casting import DatasetLike, ScorerLike
from weave.flow.dataset import Dataset
from weave.flow.model import (
    ApplyModelError,
    Model,
    PreprocessModelInput,
    apply_model_async,
)
from weave.flow.obj import Object
from weave.flow.scorer import (
    Scorer,
    _has_oldstyle_scorers,
    auto_summarize,
    get_scorer_attributes,
)
from weave.flow.util import make_memorable_name, transpose
from weave.trace.env import get_weave_parallelism
from weave.trace.objectify import register_object
from weave.trace.op import CallDisplayNameFunc, Op, OpCallError, as_op, is_op
from weave.trace.vals import WeaveObject
from weave.trace.weave_client import Call, get_ref

console = Console()
logger = logging.getLogger(__name__)

INVALID_MODEL_ERROR = (
    "`Evaluation.evaluate` requires a `Model` or `Op` instance as the `model` argument. "
    + "If you are using a function, wrap it with `weave.op` to create an `Op` instance."
)


def default_evaluation_display_name(call: Call) -> str:
    date = datetime.now().strftime("%Y-%m-%d")
    unique_name = make_memorable_name()
    return f"eval-{date}-{unique_name}"


class EvaluationResults(Object):
    rows: weave.Table


@register_object
class Evaluation(Object):
    """
    Sets up an evaluation which includes a set of scorers and a dataset.

    Calling evaluation.evaluate(model) will pass in rows from a dataset into a model matching
        the names of the columns of the dataset to the argument names in model.predict.

    Then it will call all of the scorers and save the results in weave.

    If you want to preprocess the rows from the dataset you can pass in a function
    to preprocess_model_input.

    Examples:

    ```python
    # Collect your examples
    examples = [
        {"question": "What is the capital of France?", "expected": "Paris"},
        {"question": "Who wrote 'To Kill a Mockingbird'?", "expected": "Harper Lee"},
        {"question": "What is the square root of 64?", "expected": "8"},
    ]

    # Define any custom scoring function
    @weave.op()
    def match_score1(expected: str, model_output: dict) -> dict:
        # Here is where you'd define the logic to score the model output
        return {'match': expected == model_output['generated_text']}

    @weave.op()
    def function_to_evaluate(question: str):
        # here's where you would add your LLM call and return the output
        return  {'generated_text': 'Paris'}

    # Score your examples using scoring functions
    evaluation = Evaluation(
        dataset=examples, scorers=[match_score1]
    )

    # Start tracking the evaluation
    weave.init('intro-example')
    # Run the evaluation
    asyncio.run(evaluation.evaluate(function_to_evaluate))
    ```
    """

    dataset: DatasetLike
    scorers: Optional[list[ScorerLike]] = None
    preprocess_model_input: Optional[PreprocessModelInput] = None
    trials: int = 1

    # Custom evaluation name for display in the UI.  This is the same API as passing a
    # custom `call_display_name` to `weave.op` (see that for more details).
    evaluation_name: Optional[Union[str, CallDisplayNameFunc]] = None

    # internal attr to track whether to use the new `output` or old `model_output` key for outputs
    _output_key: Literal["output", "model_output"] = PrivateAttr("output")

    @classmethod
    def from_obj(cls, obj: WeaveObject) -> Self:
        field_values = {}
        for field_name in cls.model_fields:
            if hasattr(obj, field_name):
                field_values[field_name] = getattr(obj, field_name)

        return cls(**field_values)

    def model_post_init(self, __context: Any) -> None:
        # Determine output key based on scorer types
        scorers = self.scorers or []
        if _has_oldstyle_scorers(scorers):
            self._output_key = "model_output"
            util.warn_once(
                logger,
                "Using 'model_output' key for compatibility with older scorers. Please update scorers to use 'output' parameter.",
            )

        if self.evaluation_name:
            eval_op = as_op(self.evaluate)
            eval_op.call_display_name = self.evaluation_name

        if self.name is None and self.dataset.name is not None:
            self.name = self.dataset.name + "-evaluation"  # type: ignore

    @weave.op()
    async def predict_and_score(self, model: Union[Op, Model], example: dict) -> dict:
        apply_model_result = await apply_model_async(
            model, example, self.preprocess_model_input
        )

        if isinstance(apply_model_result, ApplyModelError):
            return {
                self._output_key: None,
                "scores": {},
                "model_latency": apply_model_result.model_latency,
            }

        model_output = apply_model_result.model_output
        model_call = apply_model_result.model_call
        model_latency = apply_model_result.model_latency

        scores = {}
        if scorers := self.scorers:
            for scorer in scorers:
                apply_scorer_result = await model_call.apply_scorer(scorer, example)
                result = apply_scorer_result.result
                scorer_attributes = get_scorer_attributes(scorer)
                scorer_name = scorer_attributes.scorer_name
                scores[scorer_name] = result

        return {
            self._output_key: model_output,
            "scores": scores,
            "model_latency": model_latency,
        }

    @weave.op()
    async def summarize(self, eval_table: EvaluationResults) -> dict:
        eval_table_rows = list(eval_table.rows)
        cols = transpose(eval_table_rows)
        summary = {}

        for name, vals in cols.items():
            if name == "scores":
                if scorers := self.scorers:
                    for scorer in scorers:
                        scorer_attributes = get_scorer_attributes(scorer)
                        scorer_name = scorer_attributes.scorer_name
                        summarize_fn = scorer_attributes.summarize_fn
                        scorer_stats = transpose(vals)
                        score_table = scorer_stats[scorer_name]
                        scored = summarize_fn(score_table)
                        summary[scorer_name] = scored
            else:
                model_output_summary = auto_summarize(vals)
                if model_output_summary:
                    summary[name] = model_output_summary
        return summary

    async def get_eval_results(self, model: Union[Op, Model]) -> EvaluationResults:
        if not is_valid_model(model):
            raise ValueError(INVALID_MODEL_ERROR)
        eval_rows = []

        async def eval_example(example: dict) -> dict:
            try:
                eval_row = await self.predict_and_score(model, example)
            except OpCallError as e:
                raise e
            except Exception:
                print("Predict and score failed")
                traceback.print_exc()
                return {self._output_key: None, "scores": {}}
            return eval_row

        n_complete = 0
        dataset = self.dataset
        _rows = dataset.rows
        num_rows = len(_rows) * self.trials

        trial_rows = chain.from_iterable(repeat(_rows, self.trials))
        async for example, eval_row in util.async_foreach(
            trial_rows, eval_example, get_weave_parallelism()
        ):
            n_complete += 1
            print(f"Evaluated {n_complete} of {num_rows} examples")
            if eval_row is None:
                eval_row = {self._output_key: None, "scores": {}}
            else:
                eval_row["scores"] = eval_row.get("scores", {})
            if self.scorers:
                for scorer in self.scorers:
                    scorer_attributes = get_scorer_attributes(scorer)
                    scorer_name = scorer_attributes.scorer_name
                    if scorer_name not in eval_row["scores"]:
                        eval_row["scores"][scorer_name] = {}
            eval_rows.append(eval_row)
        return EvaluationResults(rows=weave.Table(eval_rows))

    @weave.op(call_display_name=default_evaluation_display_name)
    async def evaluate(self, model: Union[Op, Model]) -> dict:
        eval_results = await self.get_eval_results(model)
        summary = await self.summarize(eval_results)

        print("Evaluation summary", summary)

        return summary


def evaluate(
    dataset: Union[Dataset, list],
    model: Union[Op, Model],
    scorers: Optional[list[Union[Callable, Scorer]]] = None,
    preprocess_model_input: Optional[PreprocessModelInput] = None,
) -> dict:
    eval = Evaluation(
        dataset=dataset, scorers=scorers, preprocess_model_input=preprocess_model_input
    )
    return asyncio.run(eval.evaluate(model))


def is_valid_model(model: Any) -> bool:
    return (
        # Model instances are supported
        isinstance(model, Model)
        # Ops are supported
        or is_op(model)
        # Saved Models (Objects with predict) are supported
        or (
            get_ref(model) is not None
            and isinstance(model, WeaveObject)
            and hasattr(model, "predict")
            and is_op(model.predict)
        )
    )



import abc
import typing
from collections.abc import Iterator
from typing import Callable, TypeVar

from weave.trace_server import trace_server_interface as tsi
from weave.trace_server.trace_server_converter import (
    universal_ext_to_int_ref_converter,
    universal_int_to_ext_ref_converter,
)


class IdConverter:
    @abc.abstractmethod
    def ext_to_int_project_id(self, project_id: str) -> str:
        raise NotImplementedError()

    @abc.abstractmethod
    def int_to_ext_project_id(self, project_id: str) -> typing.Optional[str]:
        raise NotImplementedError()

    @abc.abstractmethod
    def ext_to_int_run_id(self, run_id: str) -> str:
        raise NotImplementedError()

    @abc.abstractmethod
    def int_to_ext_run_id(self, run_id: str) -> str:
        raise NotImplementedError()

    @abc.abstractmethod
    def ext_to_int_user_id(self, user_id: str) -> str:
        raise NotImplementedError()

    @abc.abstractmethod
    def int_to_ext_user_id(self, user_id: str) -> str:
        raise NotImplementedError()


A = TypeVar("A")
B = TypeVar("B")


class ExternalTraceServer(tsi.TraceServerInterface):
    """Used to adapt the internal trace server to the external trace server.
    This is done by converting the project_id, run_id, and user_id to their
    internal representations before calling the internal trace server and
    converting them back to their external representations before returning
    them to the caller. Additionally, we convert references to their internal
    representations before calling the internal trace server and convert them
    back to their external representations before returning them to the caller.
    """

    _internal_trace_server: tsi.TraceServerInterface
    _idc: IdConverter

    def __init__(
        self, internal_trace_server: tsi.TraceServerInterface, id_converter: IdConverter
    ):
        super().__init__()
        self._internal_trace_server = internal_trace_server
        self._idc = id_converter

    def __getattr__(self, name: str) -> typing.Any:
        return getattr(self._internal_trace_server, name)

    def _ref_apply(self, method: Callable[[A], B], req: A) -> B:
        req_conv = universal_ext_to_int_ref_converter(
            req, self._idc.ext_to_int_project_id
        )
        res = method(req_conv)
        res_conv = universal_int_to_ext_ref_converter(
            res, self._idc.int_to_ext_project_id
        )
        return res_conv

    def _stream_ref_apply(
        self, method: Callable[[A], Iterator[B]], req: A
    ) -> Iterator[B]:
        req_conv = universal_ext_to_int_ref_converter(
            req, self._idc.ext_to_int_project_id
        )
        res = method(req_conv)

        int_to_ext_project_cache = {}

        def cached_int_to_ext_project_id(project_id: str) -> typing.Optional[str]:
            if project_id not in int_to_ext_project_cache:
                int_to_ext_project_cache[project_id] = self._idc.int_to_ext_project_id(
                    project_id
                )
            return int_to_ext_project_cache[project_id]

        for item in res:
            yield universal_int_to_ext_ref_converter(item, cached_int_to_ext_project_id)

    # Standard API Below:
    def ensure_project_exists(
        self, entity: str, project: str
    ) -> tsi.EnsureProjectExistsRes:
        return self._internal_trace_server.ensure_project_exists(entity, project)

    def otel_export(self, req: tsi.OtelExportReq) -> tsi.OtelExportRes:
        req.project_id = self._idc.ext_to_int_project_id(req.project_id)
        if req.wb_user_id is not None:
            req.wb_user_id = self._idc.ext_to_int_user_id(req.wb_user_id)
        return self._ref_apply(self._internal_trace_server.otel_export, req)

    def call_start(self, req: tsi.CallStartReq) -> tsi.CallStartRes:
        req.start.project_id = self._idc.ext_to_int_project_id(req.start.project_id)
        if req.start.wb_run_id is not None:
            req.start.wb_run_id = self._idc.ext_to_int_run_id(req.start.wb_run_id)
        if req.start.wb_user_id is not None:
            req.start.wb_user_id = self._idc.ext_to_int_user_id(req.start.wb_user_id)
        return self._ref_apply(self._internal_trace_server.call_start, req)

    def call_end(self, req: tsi.CallEndReq) -> tsi.CallEndRes:
        req.end.project_id = self._idc.ext_to_int_project_id(req.end.project_id)
        return self._ref_apply(self._internal_trace_server.call_end, req)

    def call_read(self, req: tsi.CallReadReq) -> tsi.CallReadRes:
        original_project_id = req.project_id
        req.project_id = self._idc.ext_to_int_project_id(original_project_id)
        res = self._ref_apply(self._internal_trace_server.call_read, req)
        if res.call is None:
            return res
        if res.call.project_id != req.project_id:
            raise ValueError("Internal Error - Project Mismatch")
        res.call.project_id = original_project_id
        if res.call.wb_run_id is not None:
            res.call.wb_run_id = self._idc.int_to_ext_run_id(res.call.wb_run_id)
        if res.call.wb_user_id is not None:
            res.call.wb_user_id = self._idc.int_to_ext_user_id(res.call.wb_user_id)
        return res

    def calls_query(self, req: tsi.CallsQueryReq) -> tsi.CallsQueryRes:
        original_project_id = req.project_id
        req.project_id = self._idc.ext_to_int_project_id(original_project_id)
        if req.filter is not None:
            if req.filter.wb_run_ids is not None:
                req.filter.wb_run_ids = [
                    self._idc.ext_to_int_run_id(run_id)
                    for run_id in req.filter.wb_run_ids
                ]
            # TODO: How do we correctly process run_id for the query filters?
            if req.filter.wb_user_ids is not None:
                req.filter.wb_user_ids = [
                    self._idc.ext_to_int_user_id(user_id)
                    for user_id in req.filter.wb_user_ids
                ]
            # TODO: How do we correctly process user_id for the query filters?
        res = self._ref_apply(self._internal_trace_server.calls_query, req)
        for call in res.calls:
            if call.project_id != req.project_id:
                raise ValueError("Internal Error - Project Mismatch")
            call.project_id = original_project_id
            if call.wb_run_id is not None:
                call.wb_run_id = self._idc.int_to_ext_run_id(call.wb_run_id)
            if call.wb_user_id is not None:
                call.wb_user_id = self._idc.int_to_ext_user_id(call.wb_user_id)
        return res

    def calls_query_stream(self, req: tsi.CallsQueryReq) -> Iterator[tsi.CallSchema]:
        original_project_id = req.project_id
        req.project_id = self._idc.ext_to_int_project_id(original_project_id)
        if req.filter is not None:
            if req.filter.wb_run_ids is not None:
                req.filter.wb_run_ids = [
                    self._idc.ext_to_int_run_id(run_id)
                    for run_id in req.filter.wb_run_ids
                ]
            # TODO: How do we correctly process the query filters?
            if req.filter.wb_user_ids is not None:
                req.filter.wb_user_ids = [
                    self._idc.ext_to_int_user_id(user_id)
                    for user_id in req.filter.wb_user_ids
                ]
            # TODO: How do we correctly process user_id for the query filters?
        res = self._stream_ref_apply(
            self._internal_trace_server.calls_query_stream, req
        )
        for call in res:
            if call.project_id != req.project_id:
                raise ValueError("Internal Error - Project Mismatch")
            call.project_id = original_project_id
            if call.wb_run_id is not None:
                call.wb_run_id = self._idc.int_to_ext_run_id(call.wb_run_id)
            if call.wb_user_id is not None:
                call.wb_user_id = self._idc.int_to_ext_user_id(call.wb_user_id)
            yield call

    def calls_delete(self, req: tsi.CallsDeleteReq) -> tsi.CallsDeleteRes:
        req.project_id = self._idc.ext_to_int_project_id(req.project_id)
        if req.wb_user_id is not None:
            req.wb_user_id = self._idc.ext_to_int_user_id(req.wb_user_id)
        return self._ref_apply(self._internal_trace_server.calls_delete, req)

    def calls_query_stats(self, req: tsi.CallsQueryStatsReq) -> tsi.CallsQueryStatsRes:
        req.project_id = self._idc.ext_to_int_project_id(req.project_id)
        if req.filter is not None:
            if req.filter.wb_run_ids is not None:
                req.filter.wb_run_ids = [
                    self._idc.ext_to_int_run_id(run_id)
                    for run_id in req.filter.wb_run_ids
                ]
            # TODO: How do we correctly process the query filters?
            if req.filter.wb_user_ids is not None:
                req.filter.wb_user_ids = [
                    self._idc.ext_to_int_user_id(user_id)
                    for user_id in req.filter.wb_user_ids
                ]
            # TODO: How do we correctly process user_id for the query filters?
        return self._ref_apply(self._internal_trace_server.calls_query_stats, req)

    def call_update(self, req: tsi.CallUpdateReq) -> tsi.CallUpdateRes:
        req.project_id = self._idc.ext_to_int_project_id(req.project_id)
        if req.wb_user_id is not None:
            req.wb_user_id = self._idc.ext_to_int_user_id(req.wb_user_id)
        return self._ref_apply(self._internal_trace_server.call_update, req)

    def op_create(self, req: tsi.OpCreateReq) -> tsi.OpCreateRes:
        req.op_obj.project_id = self._idc.ext_to_int_project_id(req.op_obj.project_id)
        return self._ref_apply(self._internal_trace_server.op_create, req)

    def op_read(self, req: tsi.OpReadReq) -> tsi.OpReadRes:
        original_project_id = req.project_id
        req.project_id = self._idc.ext_to_int_project_id(original_project_id)
        res = self._ref_apply(self._internal_trace_server.op_read, req)
        if res.op_obj.project_id != req.project_id:
            raise ValueError("Internal Error - Project Mismatch")
        res.op_obj.project_id = original_project_id
        return res

    def ops_query(self, req: tsi.OpQueryReq) -> tsi.OpQueryRes:
        original_project_id = req.project_id
        req.project_id = self._idc.ext_to_int_project_id(original_project_id)
        res = self._ref_apply(self._internal_trace_server.ops_query, req)
        for op in res.op_objs:
            if op.project_id != req.project_id:
                raise ValueError("Internal Error - Project Mismatch")
            op.project_id = original_project_id
        return res

    def obj_create(self, req: tsi.ObjCreateReq) -> tsi.ObjCreateRes:
        req.obj.project_id = self._idc.ext_to_int_project_id(req.obj.project_id)
        if req.obj.wb_user_id is not None:
            req.obj.wb_user_id = self._idc.ext_to_int_user_id(req.obj.wb_user_id)
        return self._ref_apply(self._internal_trace_server.obj_create, req)

    def obj_read(self, req: tsi.ObjReadReq) -> tsi.ObjReadRes:
        original_project_id = req.project_id
        req.project_id = self._idc.ext_to_int_project_id(original_project_id)
        res = self._ref_apply(self._internal_trace_server.obj_read, req)
        if res.obj.project_id != req.project_id:
            raise ValueError("Internal Error - Project Mismatch")
        res.obj.project_id = original_project_id
        return res

    def objs_query(self, req: tsi.ObjQueryReq) -> tsi.ObjQueryRes:
        original_project_id = req.project_id
        req.project_id = self._idc.ext_to_int_project_id(original_project_id)
        res = self._ref_apply(self._internal_trace_server.objs_query, req)
        for obj in res.objs:
            if obj.project_id != req.project_id:
                raise ValueError("Internal Error - Project Mismatch")
            obj.project_id = original_project_id
        return res

    def obj_delete(self, req: tsi.ObjDeleteReq) -> tsi.ObjDeleteRes:
        req.project_id = self._idc.ext_to_int_project_id(req.project_id)
        return self._ref_apply(self._internal_trace_server.obj_delete, req)

    def table_create(self, req: tsi.TableCreateReq) -> tsi.TableCreateRes:
        req.table.project_id = self._idc.ext_to_int_project_id(req.table.project_id)
        return self._ref_apply(self._internal_trace_server.table_create, req)

    def table_update(self, req: tsi.TableUpdateReq) -> tsi.TableUpdateRes:
        req.project_id = self._idc.ext_to_int_project_id(req.project_id)
        return self._ref_apply(self._internal_trace_server.table_update, req)

    def table_query(self, req: tsi.TableQueryReq) -> tsi.TableQueryRes:
        req.project_id = self._idc.ext_to_int_project_id(req.project_id)
        return self._ref_apply(self._internal_trace_server.table_query, req)

    def table_query_stream(
        self, req: tsi.TableQueryReq
    ) -> Iterator[tsi.TableRowSchema]:
        req.project_id = self._idc.ext_to_int_project_id(req.project_id)
        return self._stream_ref_apply(
            self._internal_trace_server.table_query_stream, req
        )

    def table_query_stats(self, req: tsi.TableQueryStatsReq) -> tsi.TableQueryStatsRes:
        req.project_id = self._idc.ext_to_int_project_id(req.project_id)
        return self._ref_apply(self._internal_trace_server.table_query_stats, req)

    def refs_read_batch(self, req: tsi.RefsReadBatchReq) -> tsi.RefsReadBatchRes:
        return self._ref_apply(self._internal_trace_server.refs_read_batch, req)

    def file_create(self, req: tsi.FileCreateReq) -> tsi.FileCreateRes:
        req.project_id = self._idc.ext_to_int_project_id(req.project_id)
        # Special case where refs can never be part of the request
        return self._internal_trace_server.file_create(req)

    def file_content_read(self, req: tsi.FileContentReadReq) -> tsi.FileContentReadRes:
        req.project_id = self._idc.ext_to_int_project_id(req.project_id)
        # Special case where refs can never be part of the request
        return self._internal_trace_server.file_content_read(req)

    def feedback_create(self, req: tsi.FeedbackCreateReq) -> tsi.FeedbackCreateRes:
        req.project_id = self._idc.ext_to_int_project_id(req.project_id)
        original_user_id = req.wb_user_id
        if original_user_id is None:
            raise ValueError("wb_user_id cannot be None")
        req.wb_user_id = self._idc.ext_to_int_user_id(original_user_id)
        res = self._ref_apply(self._internal_trace_server.feedback_create, req)
        if res.wb_user_id != req.wb_user_id:
            raise ValueError("Internal Error - User Mismatch")
        res.wb_user_id = original_user_id
        return res

    def feedback_query(self, req: tsi.FeedbackQueryReq) -> tsi.FeedbackQueryRes:
        original_project_id = req.project_id
        req.project_id = self._idc.ext_to_int_project_id(original_project_id)
        # TODO: How to handle wb_user_id and wb_run_id in the query filters?
        res = self._ref_apply(self._internal_trace_server.feedback_query, req)
        for feedback in res.result:
            if "project_id" in feedback:
                if feedback["project_id"] != req.project_id:
                    raise ValueError("Internal Error - Project Mismatch")
                feedback["project_id"] = original_project_id
            if "wb_user_id" in feedback:
                if feedback["wb_user_id"] is not None:
                    feedback["wb_user_id"] = self._idc.int_to_ext_user_id(
                        feedback["wb_user_id"]
                    )
        return res

    def feedback_purge(self, req: tsi.FeedbackPurgeReq) -> tsi.FeedbackPurgeRes:
        req.project_id = self._idc.ext_to_int_project_id(req.project_id)
        return self._ref_apply(self._internal_trace_server.feedback_purge, req)

    def feedback_replace(self, req: tsi.FeedbackReplaceReq) -> tsi.FeedbackReplaceRes:
        req.project_id = self._idc.ext_to_int_project_id(req.project_id)
        original_user_id = req.wb_user_id
        if original_user_id is None:
            raise ValueError("wb_user_id cannot be None")
        req.wb_user_id = self._idc.ext_to_int_user_id(original_user_id)
        res = self._ref_apply(self._internal_trace_server.feedback_replace, req)
        if res.wb_user_id != req.wb_user_id:
            raise ValueError("Internal Error - User Mismatch")
        res.wb_user_id = original_user_id
        return res

    def cost_create(self, req: tsi.CostCreateReq) -> tsi.CostCreateRes:
        req.project_id = self._idc.ext_to_int_project_id(req.project_id)
        return self._ref_apply(self._internal_trace_server.cost_create, req)

    def cost_purge(self, req: tsi.CostPurgeReq) -> tsi.CostPurgeRes:
        req.project_id = self._idc.ext_to_int_project_id(req.project_id)
        return self._ref_apply(self._internal_trace_server.cost_purge, req)

    def cost_query(self, req: tsi.CostQueryReq) -> tsi.CostQueryRes:
        original_project_id = req.project_id
        req.project_id = self._idc.ext_to_int_project_id(original_project_id)
        res = self._ref_apply(self._internal_trace_server.cost_query, req)
        # Extend this to account for ORG ID when org level costs are implemented
        for cost in res.results:
            if "pricing_level_id" in cost:
                if cost["pricing_level_id"] != req.project_id:
                    raise ValueError("Internal Error - Project Mismatch")
                cost["pricing_level_id"] = original_project_id
        return res

    def actions_execute_batch(
        self, req: tsi.ActionsExecuteBatchReq
    ) -> tsi.ActionsExecuteBatchRes:
        req.project_id = self._idc.ext_to_int_project_id(req.project_id)
        original_user_id = req.wb_user_id
        if original_user_id is None:
            raise ValueError("wb_user_id cannot be None")
        req.wb_user_id = self._idc.ext_to_int_user_id(original_user_id)
        res = self._ref_apply(self._internal_trace_server.actions_execute_batch, req)
        return res

    def completions_create(
        self, req: tsi.CompletionsCreateReq
    ) -> tsi.CompletionsCreateRes:
        req.project_id = self._idc.ext_to_int_project_id(req.project_id)
        res = self._ref_apply(self._internal_trace_server.completions_create, req)
        return res



"""Settings for Weave.

## `disabled`

* Environment Variable: `WEAVE_DISABLED`
* Settings Key: `disabled`
* Default: `False`
* Type: `bool`

If True, all weave ops will behave like regular functions and no network requests will be made.

## `print_call_link`

* Environment Variable: `WEAVE_PRINT_CALL_LINK`
* Settings Key: `print_call_link`
* Default: `True`
* Type: `bool`

If True, prints a link to the Weave UI when calling a weave op.
"""

import os
from contextvars import ContextVar
from pathlib import Path
from typing import Any, Optional, Union

from pydantic import BaseModel, ConfigDict, PrivateAttr

SETTINGS_PREFIX = "WEAVE_"

# Attention Devs:
# To add new settings:
# 1. Add a new field to `UserSettings`
# 2. Add a new `should_{xyz}` function


class UserSettings(BaseModel):
    """User configuration for Weave.

    All configs can be overridden with environment variables.  The precedence is
    environment variables > `weave.trace.settings.UserSettings`."""

    disabled: bool = False
    """Toggles Weave tracing.

    If True, all weave ops will behave like regular functions.
    Can be overridden with the environment variable `WEAVE_DISABLED`"""

    print_call_link: bool = True
    """Toggles link printing to the terminal.

    If True, prints a link to the Weave UI when calling a weave op.
    Can be overridden with the environment variable `WEAVE_PRINT_CALL_LINK`"""

    capture_code: bool = True
    """Toggles code capture for ops.

    If True, saves code for ops so they can be reloaded for later use.
    Can be overridden with the environment variable `WEAVE_CAPTURE_CODE`

    WARNING: Switching between `save_code=True` and `save_code=False` mid-script
    may lead to unexpected behavior.  Make sure this is only set once at the start!
    """

    redact_pii: bool = False
    """Toggles PII redaction using Microsoft Presidio.

    If True, redacts PII from trace data before sending to the server.
    Can be overriden with the environment variable `WEAVE_REDACT_PII`
    """

    redact_pii_fields: list[str] = []
    """List of fields to redact.

    If redact_pii is True, this list of fields will be redacted.
    If redact_pii is False, this list is ignored.
    If this list is left empty, the default fields will be redacted.

    A list of supported fields can be found here: https://microsoft.github.io/presidio/supported_entities/
    Can be overriden with the environment variable `WEAVE_REDACT_PII_FIELDS`
    """

    capture_client_info: bool = True
    """Toggles capture of client information (Python version, SDK version) for ops."""

    capture_system_info: bool = True
    """Toggles capture of system information (OS name and version) for ops."""

    client_parallelism: Optional[int] = None
    """
    Sets the number of workers to use for background operations.
    If not set, automatically adjusts based on the number of cores.

    Setting this to 0 will effectively execute all background operations
    immediately in the main thread. This will not be great for performance,
    but can be useful for debugging.

    This cannot be changed after the client has been initialized.
    """

    use_server_cache: bool = False
    """
    Toggles caching of server responses, defaults to False

    If True, caches server responses to disk.
    Can be overridden with the environment variable `WEAVE_USE_SERVER_CACHE`
    """

    server_cache_size_limit: int = 1_000_000_000
    """
    Sets the size limit in bytes for the server cache, defaults to 1GB (1_000_000_000 bytes).
    Ignored if `use_server_cache` is False.

    Can be overridden with the environment variable `WEAVE_SERVER_CACHE_SIZE_LIMIT`
    """

    server_cache_dir: Optional[str] = None
    """
    Sets the directory for the server cache, defaults to None (temporary cache)
    Ignored if `use_server_cache` is False.

    Can be overridden with the environment variable `WEAVE_SERVER_CACHE_DIR`
    """

    scorers_dir: str = str(Path.home() / ".cache" / "wandb" / "weave-scorers")
    """
    Sets the directory for the scorers model checkpoints. Defaults to
    ~/.cache/wandb/weave-scorers.

    Can be overridden with the environment variable `WEAVE_SCORERS_DIR`

    """
    max_calls_queue_size: int = 100_000
    """
    Sets the maximum size of the calls queue.  Defaults to 100_000.
    Setting a value of 0 means the queue can grow unbounded.

    Can be overridden with the environment variable `WEAVE_MAX_CALLS_QUEUE_SIZE`
    """

    retry_max_interval: float = 60 * 5  # 5 min
    """
    Sets the maximum interval between retries.  Defaults to 5 minutes.

    Can be overridden with the environment variable `WEAVE_RETRY_MAX_INTERVAL`
    """

    retry_max_attempts: int = 3
    """
    Sets the maximum number of retries.  Defaults to 3.

    Can be overridden with the environment variable `WEAVE_RETRY_MAX_ATTEMPTS`
    """

    model_config = ConfigDict(extra="forbid")
    _is_first_apply: bool = PrivateAttr(True)

    def _reset(self) -> None:
        for name, field in self.model_fields.items():
            setattr(self, name, field.default)

    def apply(self) -> None:
        if self._is_first_apply:
            self._is_first_apply = False
        else:
            self._reset()

        for name in self.model_fields:
            context_var = _context_vars[name]
            context_var.set(getattr(self, name))


def should_disable_weave() -> bool:
    return _should("disabled")


def should_print_call_link() -> bool:
    return _should("print_call_link")


def should_capture_code() -> bool:
    return _should("capture_code")


def should_capture_client_info() -> bool:
    return _should("capture_client_info")


def should_capture_system_info() -> bool:
    return _should("capture_system_info")


def client_parallelism() -> Optional[int]:
    return _optional_int("client_parallelism")


def should_redact_pii() -> bool:
    return _should("redact_pii")


def redact_pii_fields() -> list[str]:
    return _list_str("redact_pii_fields")


def use_server_cache() -> bool:
    return _should("use_server_cache")


def server_cache_size_limit() -> int:
    return _optional_int("server_cache_size_limit") or 1_000_000_000


def server_cache_dir() -> Optional[str]:
    return _optional_str("server_cache_dir")


def scorers_dir() -> str:
    return _optional_str("scorers_dir")  # type: ignore


def max_calls_queue_size() -> int:
    max_queue_size = _optional_int("max_calls_queue_size")
    if max_queue_size is None:
        return 100_000
    return max_queue_size


def retry_max_attempts() -> int:
    """Returns the maximum number of retry attempts."""
    max_attempts = _optional_int("retry_max_attempts")
    if max_attempts is None:
        return 3
    return max_attempts


def retry_max_interval() -> float:
    """Returns the maximum interval between retries in seconds."""
    max_interval = _optional_float("retry_max_interval")
    if max_interval is None:
        return 60 * 5  # 5 minutes
    return max_interval


def parse_and_apply_settings(
    settings: Optional[Union[UserSettings, dict[str, Any]]] = None,
) -> None:
    if isinstance(settings, UserSettings):
        user_settings = settings
    elif isinstance(settings, dict):
        user_settings = UserSettings.model_validate(settings)
    else:
        user_settings = UserSettings()

    user_settings.apply()


_context_vars = {
    name: ContextVar(name, default=field.default)
    for name, field in UserSettings.model_fields.items()
}


def _str2bool_truthy(v: str) -> bool:
    return v.lower() in ("yes", "true", "1", "on")


def _should(name: str) -> bool:
    if env := os.getenv(f"{SETTINGS_PREFIX}{name.upper()}"):
        return _str2bool_truthy(env)
    return _context_vars[name].get()


def _optional_int(name: str) -> Optional[int]:
    if env := os.getenv(f"{SETTINGS_PREFIX}{name.upper()}"):
        return int(env)
    return _context_vars[name].get()


def _list_str(name: str) -> list[str]:
    if env := os.getenv(f"{SETTINGS_PREFIX}{name.upper()}"):
        return env.split(",")
    return _context_vars[name].get() or []


def _optional_str(name: str) -> Optional[str]:
    if env := os.getenv(f"{SETTINGS_PREFIX}{name.upper()}"):
        return env
    return _context_vars[name].get()


def _optional_float(name: str) -> Optional[float]:
    if env := os.getenv(f"{SETTINGS_PREFIX}{name.upper()}"):
        return float(env)
    return _context_vars[name].get()


__doc_spec__ = [UserSettings]



from __future__ import annotations

import asyncio
import logging
from collections.abc import Callable, Iterator
from contextlib import contextmanager
from contextvars import ContextVar
from datetime import datetime
from types import MethodType
from typing import Annotated, Any, TypeVar, Union, cast

from pydantic import (
    BaseModel,
    BeforeValidator,
    ConfigDict,
    Field,
    PrivateAttr,
    validate_call,
)

import weave
from weave.flow.dataset import Dataset
from weave.flow.eval import Evaluation, default_evaluation_display_name
from weave.flow.model import Model
from weave.flow.scorer import Scorer, auto_summarize
from weave.flow.util import make_memorable_name
from weave.trace.context import call_context
from weave.trace.context.weave_client_context import require_weave_client
from weave.trace.op import Op
from weave.trace.weave_client import Call

T = TypeVar("T")
ID = str
ScoreType = Union[float, bool, dict]

logger = logging.getLogger(__name__)

# Context variable to store the current output safely between threads.  This also
# ensures that only 1 version of the predict method is saved because the code
# contents are always the same.
current_output: ContextVar[Any] = ContextVar("current_output", default=None)
current_score: ContextVar[ScoreType | None] = ContextVar("current_score", default=None)
current_summary: ContextVar[dict | None] = ContextVar("current_summary", default=None)
current_predict_call: ContextVar[Call | None] = ContextVar(
    "current_predict_call", default=None
)


@contextmanager
def _set_current_output(output: Any) -> Iterator[None]:
    """Set the current output in a thread-safe way using context variables."""
    token = current_output.set(output)
    try:
        yield
    finally:
        current_output.reset(token)


@contextmanager
def _set_current_score(score: ScoreType) -> Iterator[None]:
    token = current_score.set(score)
    try:
        yield
    finally:
        current_score.reset(token)


@contextmanager
def _set_current_summary(summary: dict) -> Iterator[None]:
    token = current_summary.set(summary)
    try:
        yield
    finally:
        current_summary.reset(token)


def _cast_to_cls(type_: type[T]) -> Callable[[str | T], T]:
    def _convert_to_cls_inner(value: str | T) -> T:
        if isinstance(value, str):
            cls_name = value

            # Dynamically create the class if the user only provides a name
            cls_name = _validate_class_name(cls_name)

            pydantic_config_dict = {
                "__annotations__": {"name": str},
                "name": cls_name,
            }

            cls = type(cls_name, (type_,), pydantic_config_dict)
            return cast(T, cls())

        elif isinstance(value, dict):
            attributes = value

            # Dynamically create the class with attributes from the dict)
            cls_name = f"Dynamic{type_.__name__}"
            if "name" not in attributes:
                attributes["name"] = cls_name

            pydantic_config_dict = {
                "__annotations__": dict.fromkeys(attributes, Any),
                **attributes,
            }
            cls = type(cls_name, (type_,), pydantic_config_dict)
            return cast(T, cls())

        elif isinstance(value, type_):
            return value

        raise TypeError("Unsupported type for casting")

    return _convert_to_cls_inner


def _cast_to_imperative_dataset(value: Dataset | list[dict] | str) -> Dataset:
    if isinstance(value, str):
        return Dataset(name=value, rows=weave.Table([{"dataset_id": value}]))
    elif isinstance(value, list):
        return Dataset(rows=weave.Table(value))
    elif isinstance(value, Dataset):
        return value
    else:
        raise TypeError("Unsupported type for casting")


def _default_dataset_name() -> str:
    date = datetime.now().strftime("%Y-%m-%d")
    unique_name = make_memorable_name()
    return f"{date}-{unique_name}-dataset"


def _validate_class_name(name: str) -> str:
    """Validate the scorer name to be a valid class name."""
    # Check if name is not empty
    if not name:
        raise ValueError("Scorer name cannot be empty")

    # Check if name follows Python class naming conventions
    # Class names should start with a letter or underscore and contain only
    # alphanumeric characters and underscores
    import re

    if not re.match(r"^[a-zA-Z_][a-zA-Z0-9_]*$", name):
        raise ValueError(
            f"Invalid scorer name: '{name}'. Scorer names must start with a letter or underscore "
            "and contain only alphanumeric characters and underscores."
        )

    # Check if name is not a Python keyword
    import keyword

    if keyword.iskeyword(name):
        raise ValueError(f"Scorer name '{name}' cannot be a Python keyword")

    return name


class ImperativeScoreLogger(BaseModel):
    """This class provides an imperative interface for logging scores."""

    # model_id: ID
    predict_and_score_call: Call
    evaluate_call: Call
    predict_call: Call

    _captured_scores: dict[str, ScoreType] = PrivateAttr(default_factory=dict)
    _has_finished: bool = PrivateAttr(False)

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def finish(self) -> None:
        if self._has_finished:
            logger.warn("(NO-OP): Already called finish, returning.")
            return

        scores = self._captured_scores

        wc = require_weave_client()
        wc.finish_call(
            self.predict_and_score_call,
            output={
                "model_output": self.predict_call.output,
                "scores": scores,
                "model_latency": None,
            },
        )

        self._has_finished = True

    def log_score(self, scorer: Scorer | str, score: ScoreType) -> None:
        """Log a score synchronously by calling the async method.

        This is a convenience method for when you don't want to use async/await.
        If called from within an existing event loop, use _alog_score instead.
        """
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            # If there is no running loop, run the method synchronously
            asyncio.run(self._alog_score(scorer, score))
        else:
            # If there is a running loop, run the method asynchronously
            loop.create_task(self._alog_score(scorer, score))

    @validate_call
    async def _alog_score(
        self,
        scorer: Annotated[
            Scorer | dict | str,
            BeforeValidator(_cast_to_cls(Scorer)),
            Field(
                description="A metadata-only scorer used for comparisons."
                "Alternatively, you can pass a dict of attributes or just a string"
                "representing the ID of your scorer."
            ),
        ],
        score: ScoreType,
    ) -> None:
        if self._has_finished:
            raise ValueError("Cannot log score after finish has been called")

        # this is safe; pydantic casting is done in validator above
        scorer = cast(Scorer, scorer)

        @weave.op(name=scorer.name, enable_code_capture=False)
        def score_method(self: Scorer, *, output: Any, inputs: Any) -> ScoreType:
            # TODO: can't use score here because it will cause version mismatch
            # return score
            return cast(ScoreType, current_score.get())

        scorer.__dict__["score"] = MethodType(score_method, scorer)

        # attach the score feedback to the predict call
        with call_context.set_call_stack(
            [self.evaluate_call, self.predict_and_score_call]
        ):
            with _set_current_score(score):
                await self.predict_call.apply_scorer(scorer)

        # this is always true because of how the scorer is created in the validator
        scorer_name = cast(str, scorer.name)
        self._captured_scores[scorer_name] = score


class ImperativeEvaluationLogger(BaseModel):
    """This class provides an imperative interface for logging evaluations.

    An evaluation is started automatically when the first prediction is logged
    using the `log_prediction` method, and finished when the `log_summary` method
    is called.

    Each time you log a prediction, you will get back an `ImperativePredictionLogger`
    object.  You can use this object to log scores and metadata for that specific
    prediction (see that class for more details).

    Example:
        ```python
        ev = ImperativeEvaluationLogger()
        pred = ev.log_prediction(inputs, output)
        pred.log_score(scorer_name, score)
        ev.log_summary(summary)
        ```
    """

    model: Annotated[
        Model | dict | str,
        BeforeValidator(_cast_to_cls(Model)),
        Field(
            default_factory=Model,
            description="A metadata-only Model used for comparisons."
            "Alternatively, you can pass a dict of attributes or just a string"
            "representing the ID of your model.",
        ),
    ]
    dataset: Annotated[
        Dataset | list[dict] | str,
        BeforeValidator(_cast_to_imperative_dataset),
        Field(
            default_factory=lambda: Dataset(
                rows=weave.Table([{"dataset_id": _default_dataset_name()}]),
            ),
            description="A metadata-only Dataset used for comparisons."
            "If you already know your rows ahead of time, you can pass either"
            "a Dataset or list[dict]."
            "If you don't, you can just pass any string as a unique identifier",
        ),
    ]

    _eval_started: bool = PrivateAttr(False)
    _logged_summary: bool = PrivateAttr(False)
    _is_finalized: bool = PrivateAttr(False)
    _evaluate_call: Call | None = PrivateAttr(None)
    _pseudo_evaluation: Evaluation = PrivateAttr()

    # This private attr is used to keep track of predictions so we can finish
    # them if the user forgot to.
    _accumulated_predictions: list[ImperativeScoreLogger] = PrivateAttr(
        default_factory=list
    )

    def model_post_init(self, __context: Any) -> None:
        """Initialize the pseudo evaluation with the dataset from the model."""
        # At this point dataset has already been processed by the validator
        # and converted to a Dataset object
        self._pseudo_evaluation = Evaluation(
            dataset=cast(Dataset, self.dataset),
            scorers=[],
        )

        if not self._eval_started:
            self._eval_started = True

            # The following section is a "hacky" way to create Model and Evaluation
            # objects that "look right" to our object saving system.

            # --- Setup the model object ---
            @weave.op(name="Model.predict", enable_code_capture=False)
            def predict(self: Model, inputs: dict) -> Any:
                # Get the output from the context variable
                return current_output.get()

            self.model.__dict__["predict"] = MethodType(predict, self.model)

            # --- Setup the evaluation object ---
            @weave.op(name="Evaluation.evaluate", enable_code_capture=False)
            def evaluate(self: Evaluation, model: Model) -> None: ...

            @weave.op(name="Evaluation.predict_and_score", enable_code_capture=False)
            def predict_and_score(
                self: Evaluation, model: Model, example: dict
            ) -> dict:
                predict_method = cast(Op, model.get_infer_method())
                model_output, predict_call = predict_method.call(model, example)
                current_predict_call.set(predict_call)

                # This data is just a placeholder to give a sense of the data shape.
                # The actual output is explicitly replaced in ImperativeScoreLogger.finish.
                return {
                    "model_output": model_output,
                    "scores": {},
                    "model_latency": None,
                }

            @weave.op(name="Evaluation.summarize", enable_code_capture=False)
            def summarize(self: Evaluation) -> dict:
                return cast(dict, current_summary.get())

            self._pseudo_evaluation.__dict__.update(
                {
                    "evaluate": MethodType(evaluate, self._pseudo_evaluation),
                    "predict_and_score": MethodType(
                        predict_and_score, self._pseudo_evaluation
                    ),
                    "summarize": MethodType(summarize, self._pseudo_evaluation),
                }
            )

            # Create the evaluation call
            wc = require_weave_client()
            self._evaluate_call = wc.create_call(
                display_name=default_evaluation_display_name,
                op=self._pseudo_evaluation.evaluate,
                inputs={
                    "self": self._pseudo_evaluation,
                    "model": self.model,
                },
            )
            assert self._evaluate_call is not None
            call_context.push_call(self._evaluate_call)

    def _cleanup_predictions(self) -> None:
        if self._eval_started and not self._is_finalized:
            for pred in self._accumulated_predictions:
                if not pred._has_finished:
                    try:
                        pred.finish()
                    except Exception:
                        # This is best effort.  If we fail, just swallow the error.
                        pass

    def _finalize_evaluation(self, output: Any = None) -> None:
        """Handles the final steps of the evaluation: cleaning up predictions and finishing the main call."""
        if self._is_finalized:
            return

        self._cleanup_predictions()

        assert (
            self._evaluate_call is not None
        ), "Evaluation call should exist for finalization"

        # Finish the evaluation call
        wc = require_weave_client()
        # Ensure the call is finished even if there was an error during summarize or elsewhere
        try:
            wc.finish_call(self._evaluate_call, output=output)
        except Exception:
            # Log error but continue cleanup
            logger.error(
                "Failed to finish evaluation call during finalization.", exc_info=True
            )

        # Pop the call regardless of finish success
        try:
            call_context.pop_call(self._evaluate_call.id)
        except Exception:
            # If popping fails (e.g., context already unwound), log and ignore
            logger.warning("Failed to pop evaluation call from context.", exc_info=True)

        self._is_finalized = True

    def log_prediction(self, inputs: dict, output: Any) -> ImperativeScoreLogger:
        """Log a prediction to the Evaluation, and return a reference.

        The reference can be used to log scores which are attached to the specific
        prediction instance."""
        # Make the prediction call
        with _set_current_output(output):
            _, predict_and_score_call = self._pseudo_evaluation.predict_and_score.call(
                self._pseudo_evaluation,
                self.model,
                inputs,
                __require_explicit_finish=True,
            )

        # Get the predict_call from the context variable
        predict_call = current_predict_call.get()
        if predict_call is None:
            raise ValueError("predict_call should not be None")

        assert self._evaluate_call is not None
        pred = ImperativeScoreLogger(
            predict_and_score_call=predict_and_score_call,
            evaluate_call=self._evaluate_call,
            predict_call=predict_call,
        )
        self._accumulated_predictions.append(pred)
        return pred

    def log_summary(self, summary: dict | None = None) -> None:
        """Log a summary dict to the Evaluation.

        This will calculate the summary, call the summarize op, and then finalize
        the evaluation, meaning no more predictions or scores can be logged.
        """
        if self._is_finalized:
            logger.warn("(NO-OP): Evaluation already finalized, cannot log summary.")
            return

        # Calculate summary
        data_to_summarize = [
            pred._captured_scores for pred in self._accumulated_predictions
        ]
        summary_data = auto_summarize(data_to_summarize)
        final_summary = {}
        if summary_data:
            final_summary = summary_data
        if summary is not None:
            final_summary = {**final_summary, **summary}

        # Call the summarize op
        assert (
            self._evaluate_call is not None
        ), "Evaluation call should exist for summary"
        try:
            with _set_current_summary(final_summary):
                self._pseudo_evaluation.summarize()
        except Exception:
            logger.error("Error during execution of summarize op.", exc_info=True)
            # Even if summarize fails, try to finalize with the calculated summary

        self._finalize_evaluation(output=final_summary)

    def close(self) -> None:
        """Clean up the evaluation resources explicitly without logging a summary.

        Ensures all prediction calls and the main evaluation call are finalized.
        This is automatically called if the logger is used as a context manager.
        """
        if self._is_finalized:
            return

        # Finalize with None output, indicating closure without summary
        self._finalize_evaluation(output=None)

    def __del__(self) -> None:
        """Ensure cleanup happens during garbage collection."""
        if self._eval_started and not self._is_finalized:
            try:
                self.close()
            except Exception:
                # Del methods should not raise exceptions.
                # Log or handle the error appropriately if needed, but avoid propagation.
                logger.error(
                    "Error during implicit cleanup of ImperativeEvaluationLogger.",
                    exc_info=True,
                )



import json
from typing import Any

from weave.trace_server.interface.builtin_object_classes.actions import (
    ContainsWordsActionConfig,
)
from weave.trace_server.trace_server_interface import (
    CallSchema,
    TraceServerInterface,
)


def do_contains_words_action(
    project_id: str,
    config: ContainsWordsActionConfig,
    call: CallSchema,
    trace_server: TraceServerInterface,
) -> Any:
    target_words = config.target_words
    text = json.dumps(call.output)
    for word in target_words:
        if word in text:
            return True
    return False



# /// script
# requires-python = ">=3.9"
# dependencies = [
#     "weave @ git+https://github.com/wandb/weave.git@master",
# ]
# ///

# Run this script with `uv run prerelease_dry_run.py`

import datetime

import weave

# This uniq id ensures that the op is not cached
uniq_id = datetime.datetime.now().timestamp()


@weave.op
def func(a: int) -> float:
    return a + uniq_id


def main() -> None:
    client = weave.init("test-project")
    res = func(42)

    client._flush()
    calls = func.calls()

    assert len(calls) == 1
    assert calls[0].output == res
    assert calls[0].inputs == {"a": 42}


if __name__ == "__main__":
    main()
    print("Dry run passed")



import json
from typing import Any

import weave


class ValidJSONScorer(weave.Scorer):
    """Validate whether a string is valid JSON."""

    @weave.op
    def score(self, output: Any) -> dict:
        try:
            _ = json.loads(output)
        except json.JSONDecodeError:
            return {"json_valid": False}
        else:
            return {"json_valid": True}



# Clickhouse Trace Server Manager
import logging
import os
import re
from typing import Optional

from clickhouse_connect.driver.client import Client as CHClient

from weave.trace_server.costs.insert_costs import insert_costs, should_insert_costs

logger = logging.getLogger(__name__)

# These settings are only used when `replicated` mode is enabled for
# self managed clickhouse instances.
DEFAULT_REPLICATED_PATH = "/clickhouse/tables/{db}"
DEFAULT_REPLICATED_CLUSTER = "weave_cluster"


class MigrationError(RuntimeError):
    """Raised when a migration error occurs."""


class ClickHouseTraceServerMigrator:
    ch_client: CHClient
    replicated: bool
    replicated_path: str
    replicated_cluster: str

    def __init__(
        self,
        ch_client: CHClient,
        replicated: Optional[bool] = None,
        replicated_path: Optional[str] = None,
        replicated_cluster: Optional[str] = None,
    ):
        super().__init__()
        self.ch_client = ch_client
        self.replicated = False if replicated is None else replicated
        self.replicated_path = (
            DEFAULT_REPLICATED_PATH if replicated_path is None else replicated_path
        )
        self.replicated_cluster = (
            DEFAULT_REPLICATED_CLUSTER
            if replicated_cluster is None
            else replicated_cluster
        )
        self._initialize_migration_db()

    def _is_safe_identifier(self, value: str) -> bool:
        """Check if a string is safe to use as an identifier in SQL."""
        return bool(re.match(r"^[a-zA-Z0-9_\.]+$", value))

    def _format_replicated_sql(self, sql_query: str) -> str:
        """Format SQL query to use replicated engines if replicated mode is enabled."""
        if not self.replicated:
            return sql_query

        # Match "ENGINE = <optional words>MergeTree" followed by word boundary
        pattern = r"ENGINE\s*=\s*(\w+)?MergeTree\b"

        def replace_engine(match: re.Match[str]) -> str:
            engine_prefix = match.group(1) or ""
            return f"ENGINE = Replicated{engine_prefix}MergeTree"

        return re.sub(pattern, replace_engine, sql_query, flags=re.IGNORECASE)

    def _create_db_sql(self, db_name: str) -> str:
        """Geneate SQL database create string for normal and replicated databases."""
        if not self._is_safe_identifier(db_name):
            raise MigrationError(f"Invalid database name: {db_name}")

        replicated_engine = ""
        replicated_cluster = ""
        if self.replicated:
            if not self._is_safe_identifier(self.replicated_cluster):
                raise MigrationError(f"Invalid cluster name: {self.replicated_cluster}")

            replicated_path = self.replicated_path.replace("{db}", db_name)
            if not all(
                self._is_safe_identifier(part)
                for part in replicated_path.split("/")
                if part
            ):
                raise MigrationError(f"Invalid replicated path: {replicated_path}")

            replicated_cluster = f" ON CLUSTER {self.replicated_cluster}"
            replicated_engine = (
                f" ENGINE=Replicated('{replicated_path}', '{{shard}}', '{{replica}}')"
            )

        create_db_sql = f"""
            CREATE DATABASE IF NOT EXISTS {db_name}{replicated_cluster}{replicated_engine}
        """
        return create_db_sql

    def apply_migrations(
        self, target_db: str, target_version: Optional[int] = None
    ) -> None:
        status = self._get_migration_status(target_db)
        logger.info(f"""`{target_db}` migration status: {status}""")
        if status["partially_applied_version"]:
            logger.info(
                f"Unable to apply migrations to `{target_db}`. Found partially applied migration version {status['partially_applied_version']}. Please fix the database manually and try again."
            )
            return
        migration_map = self._get_migrations()
        migrations_to_apply = self._determine_migrations_to_apply(
            status["curr_version"], migration_map, target_version
        )
        if len(migrations_to_apply) == 0:
            logger.info(f"No migrations to apply to `{target_db}`")
            if should_insert_costs(status["curr_version"], target_version):
                insert_costs(self.ch_client, target_db)
            return
        logger.info(f"Migrations to apply: {migrations_to_apply}")
        if status["curr_version"] == 0:
            self.ch_client.command(self._create_db_sql(target_db))
        for target_version, migration_file in migrations_to_apply:
            self._apply_migration(target_db, target_version, migration_file)
        if should_insert_costs(status["curr_version"], target_version):
            insert_costs(self.ch_client, target_db)

    def _initialize_migration_db(self) -> None:
        self.ch_client.command(self._create_db_sql("db_management"))
        create_table_sql = """
            CREATE TABLE IF NOT EXISTS db_management.migrations
            (
                db_name String,
                curr_version UInt64,
                partially_applied_version UInt64 NULL,
            )
            ENGINE = MergeTree()
            ORDER BY (db_name)
        """
        self.ch_client.command(self._format_replicated_sql(create_table_sql))

    def _get_migration_status(self, db_name: str) -> dict:
        column_names = ["db_name", "curr_version", "partially_applied_version"]
        select_columns = ", ".join(column_names)
        query = f"""
            SELECT {select_columns} FROM db_management.migrations WHERE db_name = '{db_name}'
        """
        res = self.ch_client.query(query)
        result_rows = res.result_rows
        if res is None or len(result_rows) == 0:
            self.ch_client.insert(
                "db_management.migrations",
                data=[[db_name, 0, None]],
                column_names=column_names,
            )
        res = self.ch_client.query(query)
        result_rows = res.result_rows
        if res is None or len(result_rows) == 0:
            raise MigrationError("Migration table not found")

        return dict(zip(column_names, result_rows[0]))

    def _get_migrations(
        self,
    ) -> dict[int, dict[str, Optional[str]]]:
        migration_dir = os.path.join(os.path.dirname(__file__), "migrations")
        migration_files = os.listdir(migration_dir)
        migration_map: dict[int, dict[str, Optional[str]]] = {}
        max_version = 0
        for file in migration_files:
            if not file.endswith(".up.sql") and not file.endswith(".down.sql"):
                raise MigrationError(f"Invalid migration file: {file}")
            file_name_parts = file.split("_", 1)
            if len(file_name_parts) <= 1:
                raise MigrationError(f"Invalid migration file: {file}")
            version = int(file_name_parts[0], 10)
            if version < 1:
                raise MigrationError(f"Invalid migration file: {file}")

            is_up = file.endswith(".up.sql")

            if version not in migration_map:
                migration_map[version] = {"up": None, "down": None}

            if is_up:
                if migration_map[version]["up"] is not None:
                    raise MigrationError(
                        f"Duplicate migration file for version {version}"
                    )
                migration_map[version]["up"] = file
            else:
                if migration_map[version]["down"] is not None:
                    raise MigrationError(
                        f"Duplicate migration file for version {version}"
                    )
                migration_map[version]["down"] = file

            if version > max_version:
                max_version = version

        if len(migration_map) == 0:
            raise MigrationError("No migrations found")

        if max_version != len(migration_map):
            raise MigrationError(
                f"Invalid migration versioning. Expected {max_version} migrations but found {len(migration_map)}"
            )

        for version in range(1, max_version + 1):
            if version not in migration_map:
                raise MigrationError(f"Missing migration file for version {version}")
            if migration_map[version]["up"] is None:
                raise MigrationError(f"Missing up migration file for version {version}")
            if migration_map[version]["down"] is None:
                raise MigrationError(
                    f"Missing down migration file for version {version}"
                )

        return migration_map

    def _determine_migrations_to_apply(
        self,
        current_version: int,
        migration_map: dict,
        target_version: Optional[int] = None,
    ) -> list[tuple[int, str]]:
        if target_version is None:
            target_version = len(migration_map)
            # Do not run down migrations if not explicitly requesting target_version
            if current_version > target_version:
                logger.warning(
                    f"NOT running down migration from {current_version} to {target_version}"
                )
                return []
        if target_version < 0 or target_version > len(migration_map):
            raise MigrationError(f"Invalid target version: {target_version}")

        if target_version > current_version:
            res = []
            for i in range(current_version + 1, target_version + 1):
                if migration_map[i]["up"] is None:
                    raise MigrationError(f"Missing up migration file for version {i}")
                res.append((i, f"{migration_map[i]['up']}"))
            return res
        if target_version < current_version:
            res = []
            for i in range(current_version, target_version, -1):
                if migration_map[i]["down"] is None:
                    raise MigrationError(f"Missing down migration file for version {i}")
                res.append((i - 1, f"{migration_map[i]['down']}"))
            return res

        return []

    def _execute_migration_command(self, target_db: str, command: str) -> None:
        """Execute a single migration command in the context of the target database."""
        command = command.strip()
        if len(command) == 0:
            return
        curr_db = self.ch_client.database
        self.ch_client.database = target_db
        self.ch_client.command(self._format_replicated_sql(command))
        self.ch_client.database = curr_db

    def _update_migration_status(
        self, target_db: str, target_version: int, is_start: bool = True
    ) -> None:
        """Update the migration status in db_management.migrations table."""
        if is_start:
            self.ch_client.command(
                f"ALTER TABLE db_management.migrations UPDATE partially_applied_version = {target_version} WHERE db_name = '{target_db}'"
            )
        else:
            self.ch_client.command(
                f"ALTER TABLE db_management.migrations UPDATE curr_version = {target_version}, partially_applied_version = NULL WHERE db_name = '{target_db}'"
            )

    def _apply_migration(
        self, target_db: str, target_version: int, migration_file: str
    ) -> None:
        logger.info(f"Applying migration {migration_file} to `{target_db}`")
        migration_dir = os.path.join(os.path.dirname(__file__), "migrations")
        migration_file_path = os.path.join(migration_dir, migration_file)

        with open(migration_file_path) as f:
            migration_sql = f.read()

        # Mark migration as partially applied
        self._update_migration_status(target_db, target_version, is_start=True)

        # Execute each command in the migration
        migration_sub_commands = migration_sql.split(";")
        for command in migration_sub_commands:
            self._execute_migration_command(target_db, command)

        # Mark migration as fully applied
        self._update_migration_status(target_db, target_version, is_start=False)

        logger.info(f"Migration {migration_file} applied to `{target_db}`")



from __future__ import annotations

import dataclasses
import types
from inspect import getmro, isclass
from typing import Any, Callable

from weave.trace.op import is_op
from weave.trace_server.client_server_common.pydantic_util import (
    PydanticBaseModelGeneral,
    pydantic_asdict_one_level,
)


class ObjectRecord:
    _class_name: str
    _bases: list[str]

    def __init__(self, attrs: dict[str, Any]) -> None:
        for k, v in attrs.items():
            setattr(self, k, v)

    def __repr__(self) -> str:
        return f"ObjectRecord({self.__dict__})"

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, ObjectRecord):
            if self._class_name != other._class_name:
                return False
        else:
            if other.__class__.__name__ != getattr(self, "_class_name"):
                return False
        for k, v in self.__dict__.items():
            if k == "_class_name" or k == "_bases":
                continue
            if getattr(other, k) != v:
                return False
        return True

    def map_values(self, fn: Callable) -> ObjectRecord:
        return ObjectRecord({k: fn(v) for k, v in self.__dict__.items()})

    def unwrap(self) -> dict[str, Any]:
        # Nasty import to avoid circular import
        from weave.trace.vals import unwrap

        unwrapped_one_level = {
            k: v
            for k, v in self.__dict__.items()
            if k
            not in [
                "_class_name",
                "_bases",
                "map_values",
                "unwrap",
                "__repr__",
                "__eq__",
            ]
        }
        return unwrap(unwrapped_one_level)


def class_all_bases_names(cls: type) -> list[str]:
    # Don't include cls and don't include object
    return [c.__name__ for c in cls.mro()[1:-1]]


def pydantic_object_record(obj: PydanticBaseModelGeneral) -> ObjectRecord:
    attrs = pydantic_asdict_one_level(obj)
    for k, v in getmembers(obj, lambda x: is_op(x), lambda e: None):
        attrs[k] = types.MethodType(v, obj)
    attrs["_class_name"] = obj.__class__.__name__
    attrs["_bases"] = class_all_bases_names(obj.__class__)
    return ObjectRecord(attrs)


def dataclass_asdict_one_level(obj: Any) -> dict[str, Any]:
    if not dataclasses.is_dataclass(obj):
        raise ValueError(f"{obj} is not a dataclass")
    return {k: getattr(obj, k) for k in obj.__dataclass_fields__}


def dataclass_object_record(obj: Any) -> ObjectRecord:
    if not dataclasses.is_dataclass(obj):
        raise ValueError(f"{obj} is not a dataclass")
    attrs = dataclass_asdict_one_level(obj)
    for k, v in getmembers(obj, lambda x: is_op(x), lambda e: None):
        attrs[k] = types.MethodType(v, obj)
    attrs["_class_name"] = obj.__class__.__name__
    attrs["_bases"] = class_all_bases_names(obj.__class__)
    return ObjectRecord(attrs)


# This is an exact copy of the getmembers function from the inspect module
# with the addition of handling exceptions when calling getattr, with an on_error handler
def getmembers(
    object: Any, predicate: Any = None, on_error: Any = None
) -> list[tuple[str, Any]]:
    """Return all members of an object as (name, value) pairs sorted by name.
    Optionally, only return members that satisfy a given predicate."""
    if isclass(object):
        mro = (object,) + getmro(object)
    else:
        mro = ()
    results = []
    processed = set()
    names = dir(object)
    # :dd any DynamicClassAttributes to the list of names if object is a class;
    # this may result in duplicate entries if, for example, a virtual
    # attribute with the same name as a DynamicClassAttribute exists
    try:
        for base in object.__bases__:
            for k, v in base.__dict__.items():
                if isinstance(v, types.DynamicClassAttribute):
                    names.append(k)
    except AttributeError:
        pass
    for key in names:
        # First try to get the value via getattr.  Some descriptors don't
        # like calling their __get__ (see bug #1785), so fall back to
        # looking in the __dict__.
        try:
            value = getattr(object, key)
            # handle the duplicate key
            if key in processed:
                raise AttributeError
        except AttributeError:
            for base in mro:
                if key in base.__dict__:
                    value = base.__dict__[key]
                    break
            else:
                # could be a (currently) missing slot member, or a buggy
                # __dir__; discard and move on
                continue

        # This is where the modified code begins
        except Exception as e:
            if on_error:
                on_error(e)
            else:
                raise e
        # This is where the modified code ends

        if not predicate or predicate(value):
            results.append((key, value))
        processed.add(key)
    results.sort(key=lambda pair: pair[0])
    return results



"""The top-level functions and classes for working with Weave."""

from weave import version
from weave.trace.api import *

__version__ = version.VERSION


from weave.flow.agent import Agent as Agent
from weave.flow.agent import AgentState as AgentState
from weave.flow.annotation_spec import AnnotationSpec
from weave.flow.dataset import Dataset
from weave.flow.eval import Evaluation
from weave.flow.model import Model
from weave.flow.obj import Object
from weave.flow.prompt.prompt import EasyPrompt, MessagesPrompt, Prompt, StringPrompt
from weave.flow.scorer import Scorer
from weave.initialization import *
from weave.trace.util import Thread as Thread
from weave.trace.util import ThreadPoolExecutor as ThreadPoolExecutor
from weave.type_handlers.Markdown.markdown import Markdown

# Alias for succinct code
P = EasyPrompt

# Special object informing doc generation tooling which symbols
# to document & to associate with this module.
__docspec__ = [
    # Re-exported from trace.api
    init,
    publish,
    ref,
    require_current_call,
    get_current_call,
    finish,
    op,
    attributes,
    # Re-exported from flow module
    Object,
    Dataset,
    Model,
    Prompt,
    StringPrompt,
    MessagesPrompt,
    Evaluation,
    Scorer,
    AnnotationSpec,
    Markdown,
]



import logging
from typing import Any, Callable, Optional, TypeVar

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    ValidationInfo,
    ValidatorFunctionWrapHandler,
    model_validator,
)
from typing_extensions import Self

from weave.trace import api
from weave.trace.objectify import Objectifyable
from weave.trace.op import ObjectRef, Op
from weave.trace.ref_util import get_ref
from weave.trace.vals import WeaveObject, pydantic_getattribute

logger = logging.getLogger(__name__)

T = TypeVar("T")


def deprecated_field(new_field_name: str) -> Callable[[Callable[[Any], T]], property]:
    def decorator(func: Callable[[Any], T]) -> property:
        warning_msg = f"Use `{new_field_name}` instead of `{func.__name__}`, which is deprecated and will be removed in a future version."

        def getter(self: Any) -> T:
            logger.warning(warning_msg)
            return getattr(self, new_field_name)

        def setter(self: Any, value: T) -> None:
            logger.warning(warning_msg)
            setattr(self, new_field_name, value)

        return property(fget=getter, fset=setter)

    return decorator


class Object(BaseModel):
    name: Optional[str] = None
    description: Optional[str] = None
    ref: Optional[ObjectRef] = Field(default=None, repr=False)

    # Allow Op attributes
    model_config = ConfigDict(
        ignored_types=(Op,),
        arbitrary_types_allowed=True,
        protected_namespaces=(),
        extra="forbid",
        # Intended to be used to allow "deprecated" aliases for fields until we fully remove them.
        populate_by_name=True,
    )

    __str__ = BaseModel.__repr__

    @classmethod
    def from_uri(cls, uri: str, *, objectify: bool = True) -> Self:
        if not isinstance(cls, Objectifyable):
            raise NotImplementedError(
                f"`{cls.__name__}` must implement `from_obj` to support deserialization from a URI."
            )
        return api.ref(uri).get(objectify=objectify)

    # This is a "wrap" validator meaning we can run our own logic before
    # and after the standard pydantic validation.
    @model_validator(mode="wrap")
    @classmethod
    def handle_relocatable_object(
        cls, v: Any, handler: ValidatorFunctionWrapHandler, info: ValidationInfo
    ) -> Any:
        if isinstance(v, ObjectRef):
            return v.get()
        if isinstance(v, WeaveObject):
            # This is a relocated object, so destructure it into a dictionary
            # so pydantic can validate it.
            keys = v._val.__dict__.keys()
            fields = {}
            for k in keys:
                if k.startswith("_"):
                    continue
                val = getattr(v, k)
                fields[k] = val

            # pydantic validation will construct a new pydantic object
            def is_ignored_type(v: type) -> bool:
                return isinstance(v, cls.model_config["ignored_types"])

            allowed_fields = {k: v for k, v in fields.items() if not is_ignored_type(v)}
            new_obj = handler(allowed_fields)
            for k, kv in fields.items():
                if is_ignored_type(kv):
                    new_obj.__dict__[k] = kv

            # transfer ref to new object
            # We can't attach a ref directly to pydantic objects yet.
            # TODO: fix this. I think dedupe may make it so the user data ends up
            #    working fine, but not setting a ref here will cause the client
            #    to do extra work.
            if isinstance(v, WeaveObject):
                ref = get_ref(v)
                new_obj.__dict__["ref"] = ref
            # return new_obj

            return new_obj
        return handler(v)


# Enable ref tracking for Weave.Object
# We could try to do this on BaseModel, but we haven't proven that's safe.
# So only Weave Objects will get ref tracking behavior for now.
Object.__getattribute__ = pydantic_getattribute  # type: ignore



from enum import Enum

from pydantic import Field

from weave.trace_server.interface.builtin_object_classes import base_object_def


class ProviderReturnType(str, Enum):
    OPENAI = "openai"


class Provider(base_object_def.BaseObject):
    base_url: str
    api_key_name: str
    extra_headers: dict[str, str] = Field(default_factory=dict)
    return_type: ProviderReturnType = Field(default=ProviderReturnType.OPENAI)


class ProviderModel(base_object_def.BaseObject):
    provider: base_object_def.RefStr
    max_tokens: int



from weave.trace_server.interface.builtin_object_classes.actions import ActionSpec
from weave.trace_server.interface.builtin_object_classes.annotation_spec import (
    AnnotationSpec,
)
from weave.trace_server.interface.builtin_object_classes.base_object_def import (
    BaseObject,
)
from weave.trace_server.interface.builtin_object_classes.leaderboard import Leaderboard
from weave.trace_server.interface.builtin_object_classes.provider import (
    Provider,
    ProviderModel,
)
from weave.trace_server.interface.builtin_object_classes.saved_view import SavedView
from weave.trace_server.interface.builtin_object_classes.test_only_example import (
    TestOnlyExample,
    TestOnlyNestedBaseObject,
)

BUILTIN_OBJECT_REGISTRY: dict[str, type[BaseObject]] = {}


def register_base_object(cls: type[BaseObject]) -> None:
    """
    Register a BaseObject class in the global registry.

    Args:
        cls: The BaseObject class to register
    """
    BUILTIN_OBJECT_REGISTRY[cls.__name__] = cls


register_base_object(TestOnlyExample)
register_base_object(TestOnlyNestedBaseObject)
register_base_object(Leaderboard)
register_base_object(ActionSpec)
register_base_object(AnnotationSpec)
register_base_object(Provider)
register_base_object(ProviderModel)
register_base_object(SavedView)



from rich.console import Console
from rich.padding import Padding

console = Console()


class LogEvents:
    @staticmethod
    def step_start(name: str, color: str) -> None:
        console.rule(f"[bold {color}]Begin {name} step")

    @staticmethod
    def chat_response_start() -> None:
        pass

    @staticmethod
    def chat_message_content_delta(message_content_delta: str) -> None:
        console.print(message_content_delta, end="")

    @staticmethod
    def chat_response_complete(agent_response: str) -> None:
        console.print("\n")

    @staticmethod
    def tool_call_start(tool_call: str) -> None:
        console.print(f"[bold yellow]Tool call: [/bold yellow]{tool_call}\n")

    @staticmethod
    def tool_call_complete(tool_response: str) -> None:
        lines = tool_response.split("\n")
        if len(lines) > 4:
            lines = lines[:4]
            lines.append("...")
            tool_response = "\n".join(lines)
        console.print(
            Padding.indent(f"{tool_response}\n", 4),
            no_wrap=True,
            overflow="ellipsis",
        )

    @staticmethod
    def user_input_complete(user_input: str) -> None:
        console.print()



# This script populates the llm_token_prices table with the costs for the models that we support from costs.json
# It pulls existing costs from the table and filters out the ones that have no changes
# It then inserts the remaining costs into the table
# It is intended to run on migration
import json
import logging
import math
import os
import uuid
from datetime import datetime, timezone
from typing import Optional, TypedDict

from clickhouse_connect.driver.client import Client

COST_FILE = "cost_checkpoint.json"

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def get_current_costs(
    client: Client,
) -> list[tuple[str, float, float, datetime]]:
    current_costs = client.query(
        """
        SELECT
            llm_id,
            prompt_token_cost,
            completion_token_cost,
            effective_date
        FROM llm_token_prices
        WHERE
        created_by = 'system'
        -- There should not ever be more than 10000 default rows in the table, but just in case we limit
        LIMIT 10000
        """
    )
    return current_costs.result_rows


class CostDetails(TypedDict):
    input: float
    output: float
    provider: str
    created_at: str


def load_costs_from_json(file_name: str = COST_FILE) -> dict[str, list[CostDetails]]:
    if not os.path.isabs(file_name):
        file_path = os.path.join(os.path.dirname(__file__), file_name)
    else:
        file_path = file_name

    data = {}
    try:
        with open(file_path) as file:
            data = json.load(file)
    except json.JSONDecodeError as e:
        logger.exception("JSON decode error: %s", e)
        raise
    return data


def insert_costs_into_db(client: Client, data: dict[str, list[CostDetails]]) -> None:
    rows = []
    for llm_id, costs in data.items():
        for cost in costs:
            provider_id = cost.get("provider", "default")
            input_token_cost = cost.get("input", 0)
            output_token_cost = cost.get("output", 0)
            date_str = cost.get(
                "created_at", datetime.now().strftime("%Y-%m-%d %H:%M:%S")
            )
            created_at = datetime.strptime(date_str, "%Y-%m-%d %H:%M:%S").replace(
                tzinfo=timezone.utc
            )
            rows.append(
                (
                    str(uuid.uuid4()),
                    "default",
                    "default",
                    provider_id,
                    llm_id,
                    created_at,
                    input_token_cost,
                    "USD",
                    output_token_cost,
                    "USD",
                    "system",
                    created_at,
                ),
            )
    # Insert the data into the table
    client.insert(
        "llm_token_prices",
        rows,
        column_names=[
            "id",
            "pricing_level",
            "pricing_level_id",
            "provider_id",
            "llm_id",
            "effective_date",
            "prompt_token_cost",
            "prompt_token_cost_unit",
            "completion_token_cost",
            "completion_token_cost_unit",
            "created_by",
            "created_at",
        ],
    )


def filter_out_current_costs(
    client: Client, new_costs: dict[str, list[CostDetails]]
) -> dict[str, list[CostDetails]]:
    current_costs = get_current_costs(client)
    for (
        llm_id,
        prompt_token_cost,
        completion_token_cost,
        effective_date,
    ) in current_costs:
        if llm_id not in new_costs:
            continue
        effective_date_str = effective_date.strftime("%Y-%m-%d %H:%M:%S")
        filtered_costs = []
        for cost in new_costs[llm_id]:
            # Filter out costs that already exist in the database by comparing
            # the prompt and completion token costs with a relative tolerance
            if not (
                math.isclose(prompt_token_cost, cost["input"], rel_tol=1e-7)
                and math.isclose(completion_token_cost, cost["output"], rel_tol=1e-7)
                and effective_date_str == cost["created_at"]
            ):
                filtered_costs.append(cost)
        if len(filtered_costs) == 0:
            del new_costs[llm_id]
        else:
            new_costs[llm_id] = filtered_costs
    return new_costs


def sum_costs(data: dict[str, list[CostDetails]]) -> float:
    total_costs = 0
    for costs in data.values():
        total_costs += len(costs)
    return total_costs


def insert_costs(client: Client, target_db: str) -> None:
    client.database = target_db
    # Get costs from json
    try:
        new_costs = load_costs_from_json()
    except Exception as e:
        logger.exception("Failed to load costs from json, %s", e)
        return
    logger.info("Loaded %d costs from json", sum_costs(new_costs))

    # filter out current costs
    try:
        new_costs = filter_out_current_costs(client, new_costs)
    except Exception as e:
        logger.exception("Failed to filter out current costs, %s", e)
        return

    logger.info(
        "There are %d costs to insert, after filtering out existing costs",
        sum_costs(new_costs),
    )

    if len(new_costs) == 0:
        return

    # Attempt to insert the costs into the table
    try:
        insert_costs_into_db(client, new_costs)
    except Exception as e:
        logger.exception("Failed to insert costs into db, %s", e)
        return
    logger.info("Inserted %d costs", sum_costs(new_costs))


# We only want to insert costs if the target db version is 5 or higher
# because the costs table was added in migration 5
def should_insert_costs(
    db_curr_version: int, db_target_version: Optional[int] = None
) -> bool:
    return db_target_version is None or (
        db_target_version >= 5 and db_curr_version < db_target_version
    )
